/* bucket.cc by craig silverstein.  started 12 february 1995.
 *     Once again, this code is based on that from Andrew Goldberg.
 *
 *     The bucket is an efficient storage structure for dijkstra.
 *     It provides the same functions provided by a heap (it is
 *     in fact merely a collection of heaps -- well no, not really):
 *        o Insert: Adds an item to the given bucket heap.
 *        o Delete: Removes and item fromthe given bucket heap.
 *        o RemoveMin: takes the top (minimum) item from the next bucket heap.
 *        o Empty: true if all buckets are empty (inlined in bucket.h)
 *        o InHeap: true if given node is in a bucket (inlined in bucket.h)
 *     Possible #defines:
 *     ALLSTATS -- tells how many empty buckets are scanned, and other very
 *                 expensive calculations
 *     SMALLBUCKETS -- normally buckets in the lowest level have extent of
 *                     MinArcLen.  This keeps the extent at one.
 */

#include <stdlib.h>
#include <stdio.h>                // has stderr, etc.
#include <string.h>               // has memset
#include <limits.h>               // to check for ulong overflow
#include "bucket.h"
#include "stack.h"
#include "sp.h"
#include "assert.h"

#define NEXT(pNode)          ( (pNode)->info.bckInfo.next )
#define PREV(pNode)          ( (pNode)->info.bckInfo.prev )
#define BUCKET(pNode)        ( (Bucket *)((pNode)->info.bckInfo.bucket) )
#define DIST_TO_BASE(pNode)  ( (pNode)->info.bckInfo.distToBase )

#define OFFSET(distToBase, pLevel)      \
   ( ((distToBase) - (pLevel)->minDistToBase) >> (pLevel)->logSkip )

     // computes 2^x-1 as 2^(x-1)-1+2^(x-1).  Avoids overflow if x=32, say.
#define TWO_TO_X_LESS_1(x)   ( ((1<<(x-1))-1) | (1<<(x-1)) )

#ifdef ALLSTATS                        // expensive stats to calculate
#define EMPTY_BUCKET              statEmpty++
#define EXPANDED_NODE             statExpandedNodes++
#define EXPANDED_BUCKET           statExpandedBuckets++
#define FIRST_IN_BUCKET           statFirstInBucket++
#define LAST_IN_BUCKET            statLastInBucket++
#else
#define EMPTY_BUCKET              
#define EXPANDED_NODE             
#define EXPANDED_BUCKET           
#define FIRST_IN_BUCKET           
#define LAST_IN_BUCKET            
#endif

#ifdef SMALLBUCKETS
#define LOW_LEVEL_BUCKET_SIZE(minArcLen) ( 1 )  
#else
#define LOW_LEVEL_BUCKET_SIZE(minArcLen) ( (minArcLen) > 0 ? (minArcLen) : 1 )
#endif

//------------------------------------------------------------
// BucketBin::Init()
// BucketBin::BucketBin()
// BucketBin::~BucketBin()
//     Initialization involves mostly allocating space for the
//     buckets.  The number of buckets we need is determined by
//     the maximum arc length (which is how many buckets ahead
//     we must be able to go in one step) and the number of
//     bucket levels.  We thus figure out the number of buckets
//     at each level as well.
//        Deinitialization merely deleted the buckets.  We crash
//     if there's not enough memory.
//        Init() resets indices.  It must be done each time we
//     want to use the buckets, as opposed to Bucket(), which
//     allocates memory and need only be done once.
//------------------------------------------------------------

void BucketBin::Init()
{
   Level *pLevel;
   ulong iBucket;

   for ( pLevel = rgLevels; pLevel <= topLevel; pLevel++ )
   {
      pLevel->cNodes = 0;
      pLevel->pBucket = pLevel->rgBin;              // curr bucket is leftmost
      pLevel->minDistToBase = 0;                    // which is at the base
      pLevel->maxDistToBase = pLevel->distExtent;   // max-min = extent, always
      for ( iBucket = 0; iBucket < pLevel->cBuckets; iBucket++ )
	 pLevel->rgBin[iBucket].pNode = NULL;       // empty the buckets
   }
   minAfterWrap = topLevel->rgBin + topLevel->cBuckets;   // +infinity
   cNodesAfterWrap = 0;
   baseToZero = 0.0;                                // no wraparound yet

   statEmpty = 0;
   statExpandedNodes = 0;
   statExpandedBuckets = 0;
   statFirstInBucket = 0;
   statLastInBucket = 0;
}

BucketBin::BucketBin(ulong minArcLen, ulong maxArcLen, ulong cLevels)
{
   ulong iBucket, logcLevelSize = 1, logSkip = 0;
   Level *pLevel;

   rgLevels = new Level[cLevels];       assert(rgLevels);
   topLevel = rgLevels + cLevels-1;
   
      // we can have up to minarclen-1 different values at lowest level
   while ( (LOW_LEVEL_BUCKET_SIZE(minArcLen) >> logSkip) > 1 )
      logSkip++;                         // means 2^logSkip <= minArcLen

       // determine the number of buckets per level: it's about 
       // maxAL^(1/levels) / 2^rglogSkip[0].  *BUT* we insist it's a
       // power of two.  In addition, we need one extra bucket at the
       // top level so we don't wrap all the way around.  We actually
       // *double* the number of top-level buckets to keep powers of 2.
       // We store info in ulongs, so (# of buckets on level * skip) <= 32
   for ( logcLevelSize = 1;
	 (maxArcLen >> (logcLevelSize*cLevels + logSkip)) > 0;
	 logcLevelSize++ )
      if ( logcLevelSize*cLevels+1 + logSkip > sizeof(maxArcLen)*8 )
      {
	 fprintf(stderr, "FATAL ERROR: Maxarclen %ld is too long\n", maxArcLen);
	 exit(1);
      }

   for ( pLevel = rgLevels; pLevel <= topLevel; pLevel++ )
      pLevel->logcBuckets = logcLevelSize;
   topLevel->logcBuckets++;             // extra room at the top level
   rgLevels[0].logSkip = logSkip;       // the bootstrap value
   for ( pLevel = rgLevels; pLevel <= topLevel; pLevel++ )
   {
      pLevel->cBuckets = 1 << pLevel->logcBuckets;
      if ( pLevel != topLevel )        // note 0 has already been calculated
	 (pLevel+1)->logSkip = pLevel->logSkip + pLevel->logcBuckets;
      pLevel->distExtent = TWO_TO_X_LESS_1(pLevel->logSkip+pLevel->logcBuckets);
      pLevel->rgBin = new Bucket[pLevel->cBuckets];
      for ( iBucket = 0; iBucket < pLevel->cBuckets; iBucket++ )
	 pLevel->rgBin[iBucket].pLevel = pLevel;   // store level info
   }
   Init();
}

BucketBin::~BucketBin()
{
   Level *pLevel;

   for ( pLevel = rgLevels; pLevel <= topLevel; pLevel++ )
      delete pLevel->rgBin;
   delete rgLevels;
}

//------------------------------------------------------------
// BucketBin::DistToBucket()
//      A node's distance (value) determines whether or not it
//      belongs in a given bucket.  Every node fits in some top
//      level bucket, but depending on what lower level bucket is
//      active, a given node may not fit in any bucket in a lower
//      level.  Given a node, we return the *lowest level bucket*
//      that the node belongs in.
//         It is possible, at the top level only, to wrap around,
//      This happens when node->distToBase > top-level extent.
//      We modify the distToBase in this case (which is why a pointer
//      to it is passed).  distToBase gives the distance to
//      the leftmost top-level bucket and uniquely determines the
//      bucket a node belongs in.  It allows us to stay a ulong.
//         Check that we're not at infinity before calling.
//         RETURNS: pointer to the bucket for a given value.
//------------------------------------------------------------

Bucket *BucketBin::DistToBucket(ulong *pDistToBase)
{
   Level *pLevel;

   if ( *pDistToBase < rgLevels[0].minDistToBase )   // negative distance?  No,
      return topLevel->rgBin + OFFSET(*pDistToBase, topLevel);  // wrap-around

   for ( pLevel = rgLevels; pLevel <= topLevel; pLevel++ )
      if ( *pDistToBase <= pLevel->maxDistToBase )   // fits here
	 return pLevel->rgBin + OFFSET(*pDistToBase, pLevel);

      // For wrap-around, we only need to adjust DistToBase (take it mod)
   assert( *pDistToBase >= topLevel->distExtent+1 );   // not too small
   *pDistToBase -= topLevel->distExtent+1;             // get back in range
   assert( *pDistToBase <= topLevel->distExtent );     // not too big
   return topLevel->rgBin + OFFSET(*pDistToBase, topLevel);
}


//------------------------------------------------------------
// BucketBin::Insert()
// BucketBin::Delete()
//     This takes a node and removes it from a bucket or
//     adds it to a bucket.  We update level statistics.
//     Both insert and delete RETURN the node passed to them.
//------------------------------------------------------------

Node *BucketBin::Insert(Node *node, Bucket *bckNew)
{
   BUCKET(node) = bckNew;          // where we live, now
   if ( bckNew->pNode == NULL )    // we're the first one in the bucket
   {
      bckNew->pNode = NEXT(node) = PREV(node) = node;    // a cycle of length 1
      FIRST_IN_BUCKET;
   }
   else                            // insert us in cycle, before first
   {
      PREV(node) = PREV(bckNew->pNode);
      NEXT(node) = bckNew->pNode;
      NEXT(PREV(bckNew->pNode)) = node;
      PREV(bckNew->pNode) = node;
   }

   if ( bckNew->pLevel == topLevel && bckNew < topLevel->pBucket )
   {                                 // wrap around
      cNodesAfterWrap++;             // we adjust a special variable for wrap
      if ( bckNew < minAfterWrap )
	 minAfterWrap = bckNew;
   }
   else
      bckNew->pLevel->cNodes++;      // another not at this level
   node->where = IN_BUCKETS;
   return node;
}

Node *BucketBin::Delete(Node *node, Bucket *bckOld)
{
   assert( BUCKET(node) == bckOld );
   if ( NEXT(node) == node )    // we're the last one in the bucket
   {
      bckOld->pNode = NULL;     // (nodes loop around to form a cycle)
      LAST_IN_BUCKET;           // count a statistic
   }
   else                         // update people around us
   {
      NEXT(PREV(node)) = NEXT(node);
      PREV(NEXT(node)) = PREV(node);
      bckOld->pNode = NEXT(node);    // in case node was = pNode.  Can't hurt.
   }
   if ( bckOld->pLevel == topLevel && bckOld < topLevel->pBucket )
      cNodesAfterWrap--;             // we adjust a special variable for wrap
   else
      bckOld->pLevel->cNodes--;
   node->where = IN_NONE;       // no longer in buckets
   return node;
}

//------------------------------------------------------------
// BucketBin::SortBucket()
//     It's now time to examine a given bucket at level i.
//     Before we do this, we must sort all the buckets at
//     level i into sub-buckets at one lower level.  We
//     need to repeat this with the first non-empty sub-
//     bucket, until we've sorted all the way down to the
//     lowest level.  We're not surprised this is recursive.
//        That is, we examine the current bucket at level
//     iLevel and put its entries into bins one level lower.
//     We take the smallest of those bins and recurse.
//     RETURNS: the first bucket at the lowest level, which
//     is where we'll start examining buckets.
//------------------------------------------------------------

Bucket *BucketBin::SortBucket(Level *pLevel)
{
   Node *pNode, *nextNode;     // a node in our bucket
   Bucket *bckNew = NULL;      // the bucket we insert pNode into
   Level *pLowLevel;
   
   for ( ; pLevel > rgLevels; pLevel-- )      // tail recursion in a while loop
   {                                          // note we *don't* do level 0
      pLowLevel = pLevel-1;
      pLowLevel->minDistToBase = pLevel->minDistToBase + 
	 ( (pLevel->pBucket - pLevel->rgBin) << pLevel->logSkip );
      pLowLevel->maxDistToBase = 
	 pLowLevel->minDistToBase + pLowLevel->distExtent;
      pLowLevel->pBucket = pLowLevel->rgBin + pLowLevel->cBuckets;  // a max
      assert( pLowLevel->cNodes == 0);        // lower level should be empty
      
      pNode = pLevel->pBucket->pNode;         // the bucket we're sorting
      NEXT(PREV(pNode)) = NULL;               // turn our ring into a list
      pLevel->pBucket->pNode = NULL;          // remember to bookkeep now

      for ( ; pNode; pNode = nextNode )       // go through our list of nodes
      {
	 nextNode = NEXT(pNode);              // do this before we insert
	 bckNew = pLowLevel->rgBin + OFFSET(DIST_TO_BASE(pNode), pLowLevel);
	 assert( bckNew >= pLowLevel->rgBin && 
		 bckNew < pLowLevel->rgBin+pLowLevel->cBuckets );
	 if ( bckNew < pLowLevel->pBucket )   // set the minimum
	    pLowLevel->pBucket = bckNew;

	 Insert(pNode, bckNew);               // Insert increases cNodes
	 EXPANDED_NODE;                       // we've moved another node
      }
      assert( pLevel->cNodes >= pLowLevel->cNodes );  // how many we moved
      pLevel->cNodes -= pLowLevel->cNodes;            // these nodes have moved
      EXPANDED_BUCKET;                        // we expanded another bucket
   }
   return rgLevels[0].pBucket;                // the min bucket on bottom lvl
}

//------------------------------------------------------------
// BucketBin::RemoveMin()
//     Find the next bucket with at least one node in it and
//     returns a node from that bucket, removing it from the
//     data structure.  The hard part is when the current bucket
//     is empty; we need to advance to the next bucket on the
//     level.  It's even harder when the rest of the buckets on
//     our level is empty; we need to find the next non-empty
//     bucket on the higher level (recursively) and then sort
//     the buckets on the levels beneath it (recursively) and
//     examine those.  RETURNS: the node we've removed, or
//     NULL if there are no more nodes.
//------------------------------------------------------------

Node *BucketBin::RemoveMin()
{
   Bucket *pBucket;            // bucket holding the min node
   Level *pLevel = rgLevels;   // start at level 0

      // do bottom level specially: no need to call SortBucket here
   if ( pLevel->cNodes )                             // an item lives here
   {
      if ( pLevel->pBucket->pNode )                  // no EMPTY_BUCKET
	 return Delete(pLevel->pBucket->pNode, pLevel->pBucket);
      while ( (++(pLevel->pBucket))->pNode == NULL )
	 EMPTY_BUCKET;                               // update StatEmpty
      assert(pLevel->pBucket - pLevel->rgBin < pLevel->cBuckets);
      return Delete(pLevel->pBucket->pNode, pLevel->pBucket);
   }
   
      // do middle levels all the same.  Note that we pre-increment rgBucket
      // since we know it must be empty (it was expanded sometime before).
   for ( pLevel = rgLevels+1; pLevel <= topLevel; pLevel++ )
      if ( pLevel->cNodes )                          // an item lives here
      {
	 while ( (++(pLevel->pBucket))->pNode == NULL )
	    EMPTY_BUCKET;                            // update StatEmpty
	 assert(pLevel->pBucket - pLevel->rgBin < pLevel->cBuckets);
	 pBucket = SortBucket(pLevel);
	 assert( pBucket != NULL && pBucket->pNode != NULL );
	 return Delete(pBucket->pNode, pBucket);     // 1st node in pBucket
      }

      // didn't find anything at any level.  Now we must wrap around
   topLevel->pBucket = minAfterWrap;
   topLevel->cNodes = cNodesAfterWrap;
   minAfterWrap = topLevel->rgBin + topLevel->cBuckets;  // reinit wraps
   cNodesAfterWrap = 0;
   baseToZero += topLevel->distExtent+1;       // the base increases as we wrap

      // familiar looking: but no preincrement because minWrap might be full
   if ( topLevel->cNodes )
   {
      while ( ((topLevel->pBucket)++)->pNode == NULL )
	 EMPTY_BUCKET;                          // update StatEmpty
      (topLevel->pBucket)--;                    // undo last ++
      assert( topLevel->pBucket - topLevel->rgBin < topLevel->cBuckets );
      pBucket = SortBucket(topLevel);
      assert( pBucket != NULL && pBucket->pNode != NULL );
      return Delete(pBucket->pNode, pBucket);   // 1st node in pBucket
   }
   return NULL;                                 // never found anything
}

//------------------------------------------------------------
// BucketBin::PrintStats()
//     Prints stats specific to bucket implementation: how many
//     empty buckets we scan over, how many nodes we move from
//     one level to another, how many buckets we expand,
//     how many nodes, when put into a bucket, are the first
//     in the bucket, and how many nodes, when removed from
//     a bucket, are the last in a bucket.  We only print
//     these things when ALLSTATS is true.
//------------------------------------------------------------

void BucketBin::PrintStats()
{
#ifdef ALLSTATS
   printf("c Empty Buckets: %12ld       \n", statEmpty);
   printf("c Expanded Nodes: %11ld       Expanded Buckets: %7ld\n",
	  statExpandedNodes, statExpandedBuckets);
   printf("c First In Bucket: %10ld       Last in bucket: %9ld\n",
	  statFirstInBucket, statLastInBucket);
#endif
}

void BucketBin::dijkstra(Node *source, SP *sp)
{
   Node *currentNode, *newNode;   // newNode is beyond our current range
   Arc *arc, *lastArc;            // last arc of the current node
   Bucket *bckOld, *bckNew;

   Init();                        // reset indices
   DIST_TO_BASE(source) = (ulong)source->dist;  // no problem with wraparound
   Insert(source, DistToBucket(&(DIST_TO_BASE(source))));

   while ( (currentNode = RemoveMin()) )
   {
      sp->cScans++;
            // arcs are stored sequentially.  Our last arc is
            // one before the first arc of the next node.
      lastArc = (currentNode + 1)->first - 1;
      for ( arc = currentNode->first; arc <= lastArc; arc++ )
      {
	 newNode = arc->head;                      // where our arc ends up
	 if ( currentNode->dist + arc->len < newNode->dist )
	 {
	    bckOld = BUCKET(newNode);       // NULL if node not in a bucket
	    newNode->dist = currentNode->dist + arc->len; // we\'re shorter
	    DIST_TO_BASE(newNode) = DIST_TO_BASE(currentNode) +
	       (ulong)arc->len;
	    newNode->parent = currentNode;               // update sp tree
	    newNode->current = arc;                      // arcs and nodes
	    bckNew = DistToBucket(&(DIST_TO_BASE(newNode)));
	    
	    if ( bckOld != bckNew )             // we need to move the node
	    {
	       if ( InBucket(newNode) )         // a move, not an insert
		  Delete(newNode, bckOld);
	       Insert(newNode, bckNew);
	       sp->cUpdates++;
	    }
	 }
      }
   }
}
