/* constbucket.cc by craig silverstein.  started 1 november 1996.
 *     We try to take maximum advantage of constant-time word operations
 *     by using multi-level buckets with w buckets at each level, where
 *     w is the number of bits in a word (say, 32).  Priority queue
 *     functionality:
 *        o Insert: Adds an item to the given bucket list.
 *        o Delete: Removes and item from the given bucket list.
 *        o RemoveMin: takes the top (minimum) item from the next bucket list.
 *        o InBucket: true if node is in a bucket (inlined in constbucket.h)
 *
 *     I'll give a concrete description of what goes on for w=32.
 *     First, assume the min arc length is 1 and the max arc length is 2^32-1.
 *     The bucket a node is in is determined by the estimated path length
 *     from the node to the source.  Since no arc length > 2^32-1, we need
 *     only look at the bottom 32 bits of the path length (that is, we
 *     wrap around the path length at 2^32).  Thus, the word holding the
 *     node's path length uniquely determines its position in the bucket
 *     array.  We have w buckets per level.  Since log w = 5, we have:
 *           node->dist = xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx    (32 bits)
 *                        \/\---/\---/\---/\---/\---/\---/
 *                        ^ bucket on level 6          ^ bucket on level 0
 *     Thus we represent the position as 6 numbers (b0, b1, b2, b3, b4, b5).
 *     Suppose the last item removed was at (m0, m1, m2, m3, m4, m5).  That
 *     is the current bucket position.  If our item shares most of these
 *     values, eg (m'0, m1, m2, m3, m4, m5), then it is stored at level 0,
 *     in bucket m'0.  If it's (m'0, m'1, m'2, m3, m4, m5), then it's stored
 *     at level 2, in bucket m'2.  Only when the minimum node is in m'2 do
 *     we expand the node to level m'1 and possibly m'0.
 *        Insert works as follows: take node->dist XOR lastmin->dist.  Find
 *     the largest bit that differs.  This determines the largest m_i that
 *     differs, which tells us what bucket the node belongs in.
 *        RemoveMin works as follows: we have a bitmap at each level telling
 *     what buckets are occupied.  If the bitmap is non-zero, we advance the
 *     active bucket to be the next 1-bit, and remove it.  If the bitmap is
 *     0, we go up a level until we get to a non-zero bitmap.  We make the
 *     next bucket active, and expand it by going through all the nodes in
 *     the bucket and putting them in the proper place in the next-lower
 *     level.  We repeat with the smallest occupied bucket in each level
 *     until we get to the lowest level.  Then we remove from the lowest
 *     level as before.
 *
 *     Now suppose maxarclen = 1000 instead of 2^32-1.  We round 1000 up to
 *     2^10-1.  Now
 *        node->dist = xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx    (32 bits)
 *                                           \---/\---/
 *     and we need only two levels of buckets.  We wrap around at 2^10,
 *     not 2^32.
 *     Suppose further that minarclen = 6.  We round it down to 2^2, which
 *     means that the lowest two bits don't matter: nodes that differ only
 *     in the lowest two bits can be considered equal, since there can be
 *     no arc between them.  If maxarclen is 2^32-1, we get
 *        node->dist = xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx    (32 bits)
 *                     \---/\---/\---/\---/\---/\---/
 *     and only need 5 bucket levels.
 *
 *     Possible #defines:
 *     ALLSTATS -- tells how many empty buckets are scanned, and other very
 *                 expensive calculations
 *     SMALLBUCKETS -- normally buckets in the lowest level have extent of
 *                     MinArcLen.  This keeps the extent at one.
 */

#include <stdlib.h>
#include <stdio.h>                // has stderr, etc.
#include <string.h>               // has memset
#include <limits.h>               // to check for ulong overflow
#include "longlong.h"             // lets us find msb in constant time
#include "constbucket.h"
#include "stack.h"
#include "sp.h"
#include "assert.h"

#define NEXT(pNode)          ( (pNode)->info.constbckInfo.next )
#define PREV(pNode)          ( (pNode)->info.constbckInfo.prev )
#define LEVEL(pNode)         ( (pNode)->info.constbckInfo.iLevel )
#define BUCKET(pNode)        ( (pNode)->info.constbckInfo.iBucket )
#define DIST_TO_BASE(pNode)  ( (pNode)->info.constbckInfo.distToBase )

   // test and toggle bits corresponding to a given bucket.  The msb is bit #0!
#define BUCKET_EMPTY(iLevel, iBck) ( !(bmOccupied[iLevel] & (1<<(W-1-(iBck)))) )
#define TOGGLE_BUCKET(iLevel, iBck)  ( bmOccupied[iLevel] ^= (1<<(W-1-(iBck))) )

   // given the level, which bucket at that level a node belongs in
#define WHICH_BUCKET(pNode, iLevel) \
   ( (DIST_TO_BASE(pNode) >> shiftForPosInLevel[iLevel]) & W_MASK )

#ifdef ALLSTATS                        // expensive stats to calculate
#define EMPTY_BUCKET              statEmpty++
#define EXPANDED_NODE             statExpandedNodes++
#define EXPANDED_BUCKET           statExpandedBuckets++
#define FIRST_IN_BUCKET           statFirstInBucket++
#define LAST_IN_BUCKET            statLastInBucket++
#else
#define EMPTY_BUCKET              
#define EXPANDED_NODE             
#define EXPANDED_BUCKET           
#define FIRST_IN_BUCKET           
#define LAST_IN_BUCKET            
#endif


//------------------------------------------------------------
// ConstBucketBin::Init()
// ConstBucketBin::ConstBucketBin()
//     Initialization involves figuring out how many bucket
//     levels there needs to be, and initializing the buckets.
//     It also sets up the array that maps bit-difference-
//     between-us-and-minnode to bucket-level-we-belong-on.
//     This is affected by minarclen and maxarclen, as explained
//     in the intro description.
//        Init() clears all the buckets and resets the wraparound
//     to 0.  It must be done each time we want to use buckets
//     for a new problem (with the same min and max arc len),
//     as opposed to the constructor, which allocates memory
//     and need only be done once.
//        NB: MAX_ARC_LEN is limited to 2^(W-1) - 1, or else we
//     have trouble determining wraparound (since we have to
//     worry about overflow).
//        We don't have a destructor since arrays are statically
//     allocated (at < 1K inefficiency in space use).
//------------------------------------------------------------

void ConstBucketBin::Init()
{
   memset(bmOccupied, 0, (maxLevel+1) * sizeof(ulong)); // no bucket occupied
   distMinNode = 0;                                     // the starting distance
   currBucket = 0;                                      // dist 0 in bucket 0
   currTopBucket = 0;
   cWrapArounds = 0;                                    // no wrapping around
   statExpandedNodes = statExpandedBuckets = 0;
   statFirstInBucket = statLastInBucket = 0;
}

ConstBucketBin::ConstBucketBin(ulong minArcLen, ulong maxArcLen)
{
   int bitMinArcLen = W-1, bitMaxArcLen, bit;
   ulong zeros;              // # of zero bits before highest non-zero bit

#ifndef SMALLBUCKETS
   if ( minArcLen > 1 )      // if minArcLen is 0, then bitMinArcLen = W-1
   {
      count_leading_zeros(zeros, minArcLen);
      bitMinArcLen = zeros;  // remember, bit 0 is msb
   }
#endif
   logMinArcLen = W-1 - bitMinArcLen;

      // we want at least maxArcLen+1 room, so wraparound doesn't run into us
   count_leading_zeros(zeros, maxArcLen);
   if ( zeros == 0 )         // maxArcLen is too high   
   {
      printf("Maxarclen must be strictly less than 2^%d\n", W-1);
      exit(2);
   }
   bitMaxArcLen = zeros-1;                  // smallest power of 2 > maxArcLen
   iWrapAroundAmount = 1 << (W-1 - bitMaxArcLen);   // this is at most 2^(W-1)
   iWrapAroundAmount_1 = iWrapAroundAmount - 1;     // x & iWAA_1 == x % iWAA

   maxLevel = (bitMinArcLen - (bitMaxArcLen+1)) / LOG_W;
      // Now we set rgWhichLevel.  If our node differs from minNode in bit
      // BIT, rgWhichLevel tells us which level our node goes into.  If
      // BIT > logMinArcLen, we can go into the same bucket as minNode;
      // if BIT <= logMaxArcLen, we wrap around, so we go in the top bucket.
      // Otherwise, every LOG_W bits goes to a new level.
   for ( bit = 0; bit < W; bit++ )       // pretend bit goes from W-1 down to 0
   {
      if ( bit > bitMinArcLen )
	 rgWhichLevel[bit] = -1;         // indicates we go into currBucket
      else if ( bit <= bitMaxArcLen )    // means we wrapped around...
	 rgWhichLevel[bit] = maxLevel;   // so we go to the top level
      else
	 rgWhichLevel[bit] = (bitMinArcLen - bit) / LOG_W;
   }
      // If a node has dist d and goes in level i, how do we know
      // which bucket in the level the node goes into?  There's a series
      // of LOG_W bits in d that tells us.  shiftForPosInLevel[i] tells
      // us how much we need to shift d to the right to get to those bits.
   shiftForPosInLevel[0] = logMinArcLen;   // this is the only tricky bit
   for ( bit = 1; bit <= maxLevel; bit++ )
      shiftForPosInLevel[bit] = shiftForPosInLevel[bit-1] + LOG_W;

   Init();
}

//------------------------------------------------------------
// ConstBucketBin::DistToPosition()
//     You give me a distance as calculated from a dijkstra
//     calculation, which means it's a base path length mod
//     iWrapAroundAmount, plus some arc length.  Thus, it
//     fits into a ulong.  I take it mod iWrapAroundAmount
//     to get it back in range, and then calculate its
//     difference from distMinNode, the path length of the
//     node last removed.  The bit of the difference tells
//     us what level we're on, and the LOG_W bits starting
//     at shiftForPosInLevel tell us the bucket in the level.
//------------------------------------------------------------

void ConstBucketBin::DistToPosition(ulong *pDistToBase,
				    int *pLevel, int *pBucket)
{
   ulong diff, zeros;

   *pDistToBase &= iWrapAroundAmount_1;   // get it mod iWrapAroundAmount
   diff = *pDistToBase ^ distMinNode;
   if ( (diff >> logMinArcLen) == 0 )     // dist < minArcLen, so...
   {
      *pLevel = 0;                        // we're in the same bucket...
      *pBucket = currBucket;              // as minNode
   }
   else
   {
      count_leading_zeros(zeros, diff);   // find first bit that differs
      *pLevel = rgWhichLevel[zeros];
      assert( *pLevel != -1 );            // -1 means diff < minArcLen
      *pBucket = (*pDistToBase >> shiftForPosInLevel[*pLevel]) & W_MASK;
      assert( *pBucket >= 0 && *pBucket < W );
   }
}

//------------------------------------------------------------
// ConstBucketBin::Insert()
// ConstBucketBin::Delete()
//     This takes a node and removes it from a bucket or
//     adds it to a bucket.  If the bucket is no longer/becomes
//     empty, we toggle the bmOccupied bit instead of changing
//     pointers, since it's faster.  Thus, rgBuckets[i][j] may
//     be non-NULL for unoccupied buckets; bmOccupied is the
//     only way to determine bucket emptiness.
//     The bucket is specified by the level and the bucket
//     within the level, for insert.  Delete uses the info
//     stored with the node.
//     Both insert and delete RETURN the node passed to them.
//------------------------------------------------------------

Node *ConstBucketBin::Insert(Node *node, int iLevel, int iBucket)
{
   Node *bckNew;

   assert( iLevel >= 0 && iLevel <= maxLevel && iBucket >= 0 && iBucket < W );
   LEVEL(node) = iLevel;
   BUCKET(node) = iBucket;
   if ( BUCKET_EMPTY(iLevel, iBucket) )    // first in bucket
   {
      TOGGLE_BUCKET(iLevel, iBucket);      // indicate it's occupied
      rgBuckets[iLevel][iBucket] = NEXT(node) = PREV(node) = node;   // a cycle
      FIRST_IN_BUCKET;
   }
   else                                    // insert us in cycle, before first
   {
      bckNew = rgBuckets[iLevel][iBucket];
      PREV(node) = PREV(bckNew);
      NEXT(node) = bckNew;
      NEXT(PREV(bckNew)) = node;
      PREV(bckNew) = node;
   }
   node->where = IN_BUCKETS;
   return node;
}

Node *ConstBucketBin::Delete(Node *node)
{
   if ( NEXT(node) == node )    // we're the last one in the bucket
   {
      TOGGLE_BUCKET(LEVEL(node), BUCKET(node));           // now it's  empty
      LAST_IN_BUCKET;           // count a statistic
   }
   else                         // update people around us
   {
      NEXT(PREV(node)) = NEXT(node);
      PREV(NEXT(node)) = PREV(node);
      rgBuckets[LEVEL(node)][BUCKET(node)] = NEXT(node);  // in case rgB = node
   }
   node->where = IN_NONE;       // no longer in buckets
   return node;
}

//------------------------------------------------------------
// ConstBucketBin::ExpandLevels()
//     It's now time to examine a given bucket at level i.
//     Before we do this, we must sort all the buckets at
//     level i into sub-buckets at one lower level.  We
//     need to repeat this with the first non-empty sub-
//     bucket, until we've sorted all the way down to the
//     lowest level.  We're not surprised this is recursive.
//        That is, we examine the current bucket at level
//     iLevel and put its entries into bins one level lower.
//     We recurse on the smallest of those (occupied) bins.
//     iBucket should be the lowest non-empty bucket at iLevel.
//        Returns the first non-empty bucket at the lowest level.
//------------------------------------------------------------

int ConstBucketBin::ExpandLevels(int iLevel, int iBucket)
{
   Node *pNode, *nextNode;

   for ( ; iLevel > 0; iLevel-- )  // no need to expand at 0, the bottom level
   {                                      // this while loop is tail recursion
      assert(bmOccupied[iLevel] > 0);             // there's something to sort
      NEXT(PREV(rgBuckets[iLevel][iBucket])) = NULL;      // ring->list
      for ( pNode = rgBuckets[iLevel][iBucket]; pNode; pNode = nextNode )
      {
	 nextNode = NEXT(pNode);
	 Insert(pNode, iLevel-1, WHICH_BUCKET(pNode, iLevel-1));
	 EXPANDED_NODE;
      };
      TOGGLE_BUCKET(iLevel, iBucket);                   // bucket is now empty
      count_leading_zeros(iBucket, bmOccupied[iLevel-1]);    // for next level
      EXPANDED_BUCKET;
   }
   return iBucket;                                       // first 1 at level 0
}

//------------------------------------------------------------
// ConstBucketBin::RemoveMin()
//     Find the next bucket with at least one node in it and
//     returns a node from that bucket, removing it from the
//     data structure.  The hard part is when the current bucket
//     is empty; we need to advance to the next bucket on the
//     level.  It's even harder when the rest of the buckets on
//     our level is empty; we need to find the next non-empty
//     bucket on the higher level (recursively) and then sort
//     the buckets on the levels beneath it (recursively) and
//     examine those.  RETURNS: the node we've removed, or
//     NULL if there are no more nodes.
//------------------------------------------------------------

Node *ConstBucketBin::RemoveMin()
{
   ulong bmTop;
   int iLevel;

      // Special case: current node is not yet empty
   if ( !BUCKET_EMPTY(0, currBucket) )
      return Delete(rgBuckets[0][currBucket]);

      // First we make sure there's a bottom level bucket that's occupied
   if ( !bmOccupied[0] || maxLevel == 0 )
   {
      for ( iLevel = 1; iLevel < maxLevel; iLevel++ )    // maxLevel is special
	 if ( bmOccupied[iLevel] )
	 {
	    count_leading_zeros(currBucket, bmOccupied[iLevel]);  // temp. use
	    currBucket = ExpandLevels(iLevel, currBucket);
	    break;
	 }
            // At the top level, we may have found a bucket after wrap-around
            // (eg currbucket is bucket 8, bucket 1 is full due to wraparound,
            // but we should look at bucket 9 first).  We check this, too.
      if ( iLevel >= maxLevel )       // could be > if maxLevel = 0
	 if ( (bmOccupied[maxLevel] << currTopBucket) )  // 1 if 1-bit after curr
	 {
	    bmTop = (bmOccupied[maxLevel] << currTopBucket) >> currTopBucket;
	    count_leading_zeros(currTopBucket, bmTop);   // find next top bucket
	    currBucket = ExpandLevels(maxLevel, currTopBucket);
	 }
	 else if ( bmOccupied[maxLevel] )                // look at wrap-around
	 {
	    count_leading_zeros(currTopBucket, bmOccupied[maxLevel]);
	    currBucket = ExpandLevels(maxLevel, currTopBucket);
	    cWrapArounds++;                              // we wrapped around
	 }
	 else
	    return NULL;                                 // everything was 0s!
   }
   else
      count_leading_zeros(currBucket, bmOccupied[0]);    // level 0 is occupied

      // Now the bottom level is full, find the low bucket
   distMinNode = DIST_TO_BASE(rgBuckets[0][currBucket]); // what we'll delete
   return Delete(rgBuckets[0][currBucket]);
}


//------------------------------------------------------------
// ConstBucketBin::PrintStats()
//     Prints stats specific to bucket implementation: how many
//     empty buckets we scan over, how many nodes we move from
//     one level to another, how many buckets we expand,
//     how many nodes, when put into a bucket, are the first
//     in the bucket, and how many nodes, when removed from
//     a bucket, are the last in a bucket.  We only print
//     these things when ALLSTATS is true.
//------------------------------------------------------------

void ConstBucketBin::PrintStats()
{
#ifdef ALLSTATS
   printf("c Expanded Nodes: %11ld       Expanded Buckets: %7ld\n",
	  statExpandedNodes, statExpandedBuckets);
   printf("c First In Bucket: %10ld       Last in bucket: %9ld\n",
	  statFirstInBucket, statLastInBucket);
#endif
}

void ConstBucketBin::dijkstra(Node *source, SP *sp)
{
   Node *currentNode, *newNode;   // newNode is beyond our current range
   Arc *arc, *lastArc;            // last arc of the current node
   int iLevel, iBucket;

   Init();                        // reset indices
   DIST_TO_BASE(source) = (ulong)source->dist;  // no problem with wraparound
   DistToPosition(&(DIST_TO_BASE(source)), &iLevel, &iBucket);
   Insert(source, iLevel, iBucket);

   while ( (currentNode = RemoveMin()) )
   {
      sp->cScans++;
            // arcs are stored sequentially.  Our last arc is
            // one before the first arc of the next node.
      lastArc = (currentNode + 1)->first - 1;
      for ( arc = currentNode->first; arc <= lastArc; arc++ )
      {
	 newNode = arc->head;                      // where our arc ends up
	 if ( currentNode->dist + arc->len < newNode->dist )
	 {
	    newNode->dist = currentNode->dist + arc->len; // we\'re shorter
	    DIST_TO_BASE(newNode) = DIST_TO_BASE(currentNode) +
	       (ulong)arc->len;             // no problem with overflow
	    newNode->parent = currentNode;               // update sp tree
	    newNode->current = arc;                      // arcs and nodes
	    DistToPosition(&(DIST_TO_BASE(newNode)), &iLevel, &iBucket);
	    if ( LEVEL(newNode) != iLevel || BUCKET(newNode) != iBucket )
	    {                                   // we need to move the node
	       if ( InBucket(newNode) )         // a move, not an insert
		  Delete(newNode);
	       Insert(newNode, iLevel, iBucket);
	       sp->cUpdates++;
	    }
	 }
      }
   }
}
