/* gor shortest-path algorithm */
/* Stanford University, Computer Science Department */
/* Andrew V. Goldberg - goldberg@cs.stanford.edu */
/* modified by craig silverstein.  started 11 februrary 1995.
 * took this out of cycle-cancelling min-flow code.  Tried to
 * rewrite it to improve modularity.
 *
 *     GOR takes a graph (of the form n -- number of nodes -- and
 *     an array of nodes) and a source, and determines the shortest
 *     path tree for the graph using a topological sort and bellman-
 *     ford. 
 *     The function gor(n, nodes, source, [stkSources]), where
 *           N: number of nodes
 *           NODES: pointer to first node in the node list
 *           SOURCE: pointer to the source node
 *           STKSOURCES: stack of source nodes.  set SOURCE=NULL to use.
 *     The function returns the number of iterations of the algorithm,
 *     that is, the number of times a topolgical sort is done plus the
 *     the number of nodes looked at by bellman-ford.  This is an estimate
 *     of the amount of work done, assuming each topological sort is about
 *     the same amount of work (true if many costs must be updated).
 *        The algorithm also modifies the "dist" (distance to source) and 
 *     "parent" (parent in shortest path tree) parameters of the nodes in
 *     the graph.
 *        GOR takes an array of Nodes.  Nodes are expected to
 *     be defined in gor.h and have the following fields:
 *           long value -- holds the distance
 *           long temp  -- used internally by the dijkstra algorithm
 *           node *parent -- holds the shortest-path tree when we finish
 *           arc *current -- arcs in sp-tree (p->current is arc w/ \'rent)
 */

#include <stdlib.h>
#include "sp.h"

#define NOT_IN_STK      0    // init value, also mindist definitely assigned
#define IN_TOPSORT_STK  1    // currently we're updating distances
#define IN_SOURCES_STK  2    // we need to update distances of kids
#define IN_DFS          3    // currently topologically sorting subtree

#define STATUS(pNode)   ( (pNode)->info.status )


//-------------------------------------------------------------
// depth_first_search()
//     As part of our topological sort, we put all the nodes we
//     can reach from a given node on the stack.  We put children
//     of a node on before the node itself; that is, we do a depth-
//     first search, putting nodes on the stack as we come back up
//     from the recursion.  We take a root and a stack and modify
//     the stack.  Each node is marked IN_DFS as it is encountered
//     on the way down (to protect against cycles) and IN_STACK on
//     the way up, so we can tell where a node is.
//        A node is considered if it exists in the residual graph
//     and its dist needs to be updated.  When the shortest path
//     has been found, dfs will not modify the stack.
//        RETURNS: 1 if any child is explored, 0 if we are a dead
//     end.
//-------------------------------------------------------------

int depth_first_search(Node *pRoot, Stack *stk)
{
   Node *nodeChild;
   Arc *pArc, *arcLast;
   int retval = 0;

   STATUS(pRoot) = IN_DFS;

         // arcs are stored sequentially.  Our last arc is
         // one before the first arc of the next node.
   arcLast = (pRoot+1)->first - 1;
   for ( pArc = pRoot->first; pArc <= arcLast; pArc++ )
   {
      nodeChild = pArc->head;
      if ( pRoot->dist + pArc->len <= nodeChild->dist       // could be SP
	  && STATUS(nodeChild) != IN_DFS                    // cycle
	  && STATUS(nodeChild) != IN_TOPSORT_STK )          // cycle
      {
	 depth_first_search(nodeChild, stk);         // we like recursion
	 retval = 1;                                 // and we got somewhere
      }
   }
   
   stk->Push(pRoot);                      // done pushing kids
   STATUS(pRoot) = IN_TOPSORT_STK;
   pRoot->where = IN_GOR;
   return retval;
}

//-------------------------------------------------------------
// topological_sort()
//     We want to topologically sort that part of the graph
//     that is reachable from the sources.  We take the sources
//     from the "sources" stack and do a depth-first search on
//     them to find reachable nodes; the search puts them on
//     the "pass" stack of all nodes found in this pass.  We count
//     a node as reachable if we can reach it from the source and
//     lower its cost.
//-------------------------------------------------------------

void toplogical_sort(Stack *stkSources, Stack *stkTopSort)
{
   Node *pSource;
   Arc *pArc, *arcLast;

   while ( !stkSources->IsEmpty() )
   {
      pSource = (Node *) stkSources->Pop();
      assert( STATUS(pSource) != IN_DFS );
      if ( STATUS(pSource) == IN_SOURCES_STK )
      {
            // if we have any admissible children, we stick this tree onto
            // the topological sort stack.  If not, we take ourself out.

            // arcs are stored sequentially.  Our last arc is
            // one before the first arc of the next node.
	 arcLast = (pSource+1)->first - 1;
	 for ( pArc = pSource->first; pArc <= arcLast; pArc++ )
	    if ( pSource->dist + pArc->len < pArc->head->dist )
	    {
	       depth_first_search(pSource, stkTopSort);       // got a good kid
	       break;
	    }
	 if ( pArc > arcLast )                                // no good kid
	    STATUS(pSource) = NOT_IN_STK;
      }      // end "if"
   }         // end "while stack"
}

//-------------------------------------------------------------
// bellman_ford()
//     Bellman-ford is a shortest path algorithm.  We help it
//     be fast by giving it the nodes in topological order;
//     that is, we take the nodes off of stkTopSort.  If the
//     dist of a node is lowered, good.  If it is lowered twice,
//     that means that two sources could get to it, and the second
//     was better than the first.  We lower the dist again, but then
//     have to rescan our node\'s children, so we add it back onto
//     the sources stack, treating it as a new source of dist changed.
//        RETURNS: number of nodes we\'ve looked at
//-------------------------------------------------------------

long bellman_ford(Stack *stkTopSort, Stack *stkSources)
{
   Node *pNode, *pChild;
   Arc *pArc, *arcLast;
   long cUpdates = 0;                   // retval

   while ( !stkTopSort->IsEmpty() )     // examine edges in topological order
   {
      pNode = (Node *)stkTopSort->Pop();
      STATUS(pNode) = NOT_IN_STK;
      pNode->where = IN_NONE;           // no longer in GOR data structures

            // arcs are stored sequentially.  Our last arc is
            // one before the first arc of the next node.
      arcLast = (pNode+1)->first - 1;
      for ( pArc = pNode->first; pArc <= arcLast; pArc++ )
      {
	 pChild = pArc->head;        // where our arc is going
	 if ( pNode->dist + pArc->len < pChild->dist )
	 {
	    cUpdates++;
	    pChild->dist = pNode->dist + pArc->len;      // lower dist
	    pChild->parent = pNode;                      // update sp tree
	    pChild->current = pArc;                      // arcs and nodes
	    
	    if ( STATUS(pChild) == NOT_IN_STK )          // 2nd lowering
	    {
	       STATUS(pChild) = IN_SOURCES_STK;          // redo those below
	       stkSources->Push(pChild);
	    }
	 }              // end "if lower dist"
      }                 // end "if available edge"
   }                    // end "while stack"
   return cUpdates;
}

//-------------------------------------------------------------
// SP::gor()
//     Given the number of nodes, a list of nodes, and a source node,
//     finds the shortest path tree.  Modifies the dist (value) field
//     of each node and the parent pointer.
//        If StopAt is non-NULL, we call it on every node we encounter.
//     If any passes the test, we return.  This can be used for
//     2-point shortest path, even when the 2nd point is defined by
//     criteria rather than explicitely.  Because of the way GOR
//     works (ie bellman-ford) we have to finish the sp before
//     finding a stop-at node.  Thus this is not very efficient.
//-------------------------------------------------------------

void SP::gor(Node *source)
{
   STATUS(source) = IN_SOURCES_STK;
   stkSources->Push(source);   // now in our copy of the stack
   source->where = IN_GOR;

   while ( !stkSources->IsEmpty() )
   {
      cScans++;
      toplogical_sort(stkSources, stkTopSort);
      cUpdates += bellman_ford(stkTopSort, stkSources);
   }

}

