/* heap.c  by craig silverstein.  started 10 january 1995
 *    This is hardly my code.  I just cleaned it up a lot and
 *    made it more modular.  Basic code I got from andrew goldberg
 *    but I'm not sure if he wrote it or not.
 *    I think I'll make this a class.  Operations:
 *        o MoveUp: takes an elt in heap whose value may have changed
 *                  (shrunk) and moves it to its proper place
 *        o Insert: Adds an item to the heap.  Call MoveUp next.
 *        o RemoveMin: takes the top (minimum) item from the heap.
 *        o IsEmpty: true if heap is empty (inlined in heap.h)
 *        o InHeap: true if given node is in the heap (inlined in heap.h)
 *   The heap stores nodes.  heap.h should define or include a definition
 *   of what a node is.  We need at least the following fields:
 *        long value:    compares nodes
 *        long temp:     used internally to store where a node is in the heap
 */

#include <stdio.h>
#include "heap.h"
#include "stack.h"
#include "sp.h"
#include "assert.h"

#define LOG_HEAP_DEGREE  2       // a 2^(LHD)-ary tree

#define MIN(a,b)        ( (a) > (b) ? (b) : (a) )
#define PLACE_NODE(node, place)    \
     data[place] = node,  NODE_POS(node) = place

#ifdef ALLSTATS
#define HEAP_OP      statOps++   // take the time to mark heap operations?
#define HEAP_SWAPS   statSwaps++
#else
#define HEAP_OP      
#define HEAP_SWAPS   
#endif


//--------------------------------------------------------------
// Heap::Heap()
// Heap::~Heap()
//     The constructor takes a heap size and pre-initializes an
//     array to be that size.  That is, you must know the max
//     heap size in advance.
//        The destructor merely deallocates the array.
//--------------------------------------------------------------

Heap::Heap(long maxSizeGiven)
{
   maxSize = maxSizeGiven;         // we can error check if we want
   size = 0;                       // no nodes in here yet
   data = new Node * [maxSize];    // array of pointers to nodes
   statOps = 0;
}

Heap::~Heap()
{
   delete data;
}

//--------------------------------------------------------------
// Heap::MoveUp()
// Heap::MoveDown()
// Heap::Heapify()
//     Takes a pointer to a node whose value may have shrunk
//     and repeatedly replaces it with its parent until it\'s
//     in the proper place on the heap.
//     Takes a node, whose value has supposedly changed.
//        Heapify() moves up all the nodes so the entire heap
//     is put in heap order at once. It does this by sifting
//     items down, thus doing its work in linear time.
//--------------------------------------------------------------

void Heap::MoveUp(Node *node)
{
   long iCurrent, iParent;
   Node *nodeParent;
   
   for ( iCurrent=NODE_POS(node), iParent=(NODE_POS(node)-1) >> LOG_HEAP_DEGREE;
	 iCurrent > 0;
	 iCurrent = iParent, iParent = (iParent-1) >> LOG_HEAP_DEGREE )
   {
      HEAP_SWAPS;
      nodeParent = data[iParent];
      if ( node->dist >= nodeParent->dist )   // we\'re in the right place
	 break;
      PLACE_NODE( nodeParent, iCurrent );     // else move the parent down
   }
   PLACE_NODE( node, iCurrent );              // put the node in the hole
   HEAP_OP;
}

void Heap::MoveDown(Node *pNode)
{
   long iCurrent, iChild, iGoodChild, iFirstChild, iLastChild;
   double minValue = 0;         // min value of children of us (dbl for prec.)
   
   for ( iCurrent = NODE_POS(pNode); 
	 /* break when value <= min */;
	 iCurrent = iGoodChild )
   {
      HEAP_SWAPS;
      iFirstChild = (iCurrent << LOG_HEAP_DEGREE) + 1;
      iLastChild = MIN( iFirstChild + (1<<LOG_HEAP_DEGREE) - 1, size-1 );
      if ( iFirstChild >= size )                 // at the end of the tree
	 break;
      minValue = data[iFirstChild]->dist;
      iGoodChild = iFirstChild;                  // the child with the min
      for ( iChild = iFirstChild+1; iChild <= iLastChild; iChild++ )
	 if ( data[iChild]->dist < minValue )    // new min
	 {
	    minValue = data[iChild]->dist;
	    iGoodChild = iChild;
	 }
      if ( pNode->dist <= minValue )             // no need to move down-er
	 break;
      PLACE_NODE( data[iGoodChild], iCurrent );
   }                                             // end for
   PLACE_NODE( pNode, iCurrent );                // pNode goes in hole
   HEAP_OP;
}

void Heap::Heapify()
{
   long iNode;

         // note that most nodes are leaves, which we can ignore
    for ( iNode = (size-2) >> LOG_HEAP_DEGREE; iNode >= 0; iNode-- )
       MoveDown(data[iNode]);
}

//--------------------------------------------------------------
// Heap::Insert()
//     Inserts an element in the heap, *DOES NOT PRESERVE HEAP
//     ORDER* --  call MoveUp() to do this.  We merely insert
//     onto the end.
//--------------------------------------------------------------

void Heap::Insert(Node *node)
{
   PLACE_NODE( node, size );        // can\'t do ++ within a macro
   node->where = IN_HEAP;
   size++;
   assert( size <= maxSize );       // sanity check
}

//--------------------------------------------------------------
// Heap::FindMin()         [in heap.h]
// Heap::RemoveArbitrary()
// Heap::RemoveMin()
//     Takes the top element off the heap, puts the last element
//     in its place, and then reheapifies by moving the last
//     element down to its proper place.  Returns the top element.
//     RemoveArbitrary() removes an arbitrary member of the heap.
//--------------------------------------------------------------

Node *Heap::RemoveArbitrary()
{
   assert( size > 0 );
   size--;
   data[size]->where = IN_NONE;
   return data[size];
}

Node *Heap::RemoveMin()
{
   Node *retval;
   
   retval = data[0];            // take it from the top
   retval->where = IN_NONE;     // no longer in the heap
   size--;                      // let the heap know

   if ( size > 0 )
   {
      PLACE_NODE(data[size], 0);    // put last node at the top
      MoveDown(data[0]);            // and move it down
   }
   return retval;
}

//--------------------------------------------------------------
// Heap::Print()
//     Prints the heap as follows: breadth-first listing of the
//     nodes, each node on a line by itself.  Following the nodes
//     on the line are the values of the node\'s children.  Values
//     are all that are printed.  In this way it is easy to test
//     whether the heap property is being preserved.
//--------------------------------------------------------------

void Heap::Print()
{
   long iNode, iChild;
   
   fprintf(stderr, "\nTHE HEAP: size %li of %li", size, maxSize);
   for ( iNode = 0; iNode < size; iNode++ )
   {
      fprintf(stderr, "\nPos %li.  %f: ", iNode, data[iNode]->dist);
      for ( iChild = (iNode << LOG_HEAP_DEGREE) + 1;
	    iChild <= (iNode+1) << LOG_HEAP_DEGREE  &&  iChild < size;
            iChild++ )
      {
	 fprintf(stderr, "%f ", data[iChild]->dist);
	 if ( data[iChild]->dist < data[iNode]->dist )   // bad bad bad
	    fprintf(stderr, "** ");
      }
   }
}

//--------------------------------------------------------------------
// Heap::PrintStats()
//     Print stats specific to the heap implementation:.  Right now
//     it's pretty redundant, but useful.
//--------------------------------------------------------------------

void Heap::PrintStats()
{
#ifdef ALLSTATS
   printf("c Heap operations: %10ld       Heap swaps: %13ld\n", 
	  statOps, statSwaps);
#endif
}

//--------------------------------------------------------------------
// Heap::dijkstra_heap()
//     We find the node with the smallest distance by taking it off
//     the heap.  We update all nodes around it and put back on the
//     heap those whose distances were changed.  We start with only
//     the source on the heap and end when the heap is empty.
//        If StopAt is non-NULL, we call it on every node we encounter.
//     If any passes the test, we return.  This can be used for
//     2-point shortest path, even when the 2nd point is defined by
//     criteria rather than explicitely.
//--------------------------------------------------------------------

void Heap::dijkstra(Node *source, SP *sp)
{
   Node *currentNode, *newNode;   // newNode is beyond our current range
   Arc *arc, *lastArc;            // last arc of the current node

   Empty();                       // get the heap ready
   Insert(source);
   while ( !IsEmpty() )
   {
      currentNode = RemoveMin();                  // currentNode is optimal
      sp->cScans++;
            // arcs are stored sequentially.  Our last arc is
            // one before the first arc of the next node.
      lastArc = (currentNode + 1)->first - 1;
      for ( arc = currentNode->first; arc <= lastArc; arc++ )
      {
	 newNode = arc->head;     // where our arc ends up
	 if ( (currentNode->dist + arc->len < newNode->dist) )
	 {
	    newNode->dist = currentNode->dist + arc->len; // we're shorter
	    newNode->parent = currentNode;                // update sp tree
	    newNode->current = arc;                       // arcs and nodes
	    if ( !InHeap(newNode) )                       // virgin node
	       Insert(newNode);
	    MoveUp(newNode);                              // update anyways
	    sp->cUpdates++;
	 }
      }
   }
}
