/* hotqueue.cc  by craig silverstein.  Started 20 february 1996.
 *     See hotqueue.h for the idea.  We do bucket stuff, but keep track
 *     of how big a bucket is and use a heap, or even a sorted list, if
 *     it's small.
 */

#include <stdlib.h>
#include <stdio.h>                // has stderr, etc.
#include <string.h>               // has memset
#include <limits.h>               // to check for ulong overflow
#include <assert.h>
#include "hotqueue.h"
#include "stack.h"
#include "sp.h"

#define MIN_MIN_EXPAND       5    // never expand anything smaller than this

#define NEXT(pNode)          ( (pNode)->info.bckInfo.next )
#define PREV(pNode)          ( (pNode)->info.bckInfo.prev )
#define BUCKET(pNode)        ( (HOTBucket *)((pNode)->info.bckInfo.bucket) )
#define DIST_TO_BASE(pNode)  ( (pNode)->info.bckInfo.distToBase )

#define OFFSET(distToBase, pLevel)      \
   ( ((distToBase) - (pLevel)->minDistToBase) >> (pLevel)->logSkip )

     // computes 2^x-1 as 2^(x-1)-1+2^(x-1).  Avoids overflow if x=32, say.
#define TWO_TO_X_LESS_1(x)   ( ((1<<(x-1))-1) | (1<<(x-1)) )

#ifdef ALLSTATS                        // expensive stats to calculate
#define EMPTY_BUCKET              statEmpty++
#define EXPANDED_NODE             statExpandedNodes++
#define EXPANDED_BUCKET           statExpandedBuckets++
#define FIRST_IN_BUCKET           statFirstInBucket++
#define LAST_IN_BUCKET            statLastInBucket++
#define MOVEUP_HEAP               statMoveupHeap++
#define MOVEUP_SORT               statMoveupSort++
#define INSERT_HEAP               statInsertHeap++
#define INSERT_SORT               statInsertSort++
#define STATE_CHANGE              statStateChange++
#else
#define EMPTY_BUCKET              ;
#define EXPANDED_NODE             ;
#define EXPANDED_BUCKET           ;
#define FIRST_IN_BUCKET           ;
#define LAST_IN_BUCKET            ;
#define MOVEUP_HEAP               ;
#define MOVEUP_SORT               ;
#define INSERT_HEAP               ;
#define INSERT_SORT               ;
#define STATE_CHANGE              ;
#endif

#ifdef SMALLBUCKETS
#define LOW_LEVEL_BUCKET_SIZE(minArcLen) ( 1 )  
#else
#define LOW_LEVEL_BUCKET_SIZE(minArcLen) ( (minArcLen) > 0 ? (minArcLen) : 1 )
#endif


//------------------------------------------------------------
// HOTQueue::Init()
// HOTQueue::HOTQueue()
// HOTQueue::~HOTQueue()
//     Initialization involves mostly allocating space for the
//     buckets.  The number of buckets we need is determined by
//     the maximum arc length (which is how many buckets ahead
//     we must be able to go in one step) and the number of
//     bucket levels.  We thus figure out the number of buckets
//     at each level as well.
//        Deinitialization merely deleted the buckets.  We crash
//     if there's not enough memory.
//        Init() resets indices.  It must be done each time we
//     want to use the buckets, as opposed to Bucket(), which
//     allocates memory and need only be done once.
//------------------------------------------------------------

void HOTQueue::Init()
{
   HOTLevel *pLevel;
   ulong iBucket;

   for ( pLevel = rgLevels; pLevel <= topLevel; pLevel++ )
   {
      pLevel->cNodes = 0;
      pLevel->pBucket = pLevel->rgBin;              // curr bucket is leftmost
      pLevel->minDistToBase = 0;                    // which is at the base
      pLevel->maxDistToBase = pLevel->distExtent;   // max-min = extent, always
      for ( iBucket = 0; iBucket < pLevel->cBuckets; iBucket++ )
      {
	 pLevel->rgBin[iBucket].pNode = NULL;       // empty the buckets
	 pLevel->rgBin[iBucket].cNodes = 0;
      }
   }
   minAfterWrap = topLevel->rgBin + topLevel->cBuckets;   // +infinity
   cNodesAfterWrap = 0;
   baseToZero = 0.0;                                // no wraparound yet
   bckCurrent = topLevel->rgBin;                    // the very first bucket
   sortCurrent = NULL;                              // bckCurrent not sorted
   sortCurrentLast = NULL;
   heapCurrent->Empty();                            // remove contents
   stateCurrent = CURR_BCK;
   iSort = 0;                                       // no sort ops yet

   statEmpty = 0;
   statExpandedNodes = 0;
   statExpandedBuckets = 0;
   statFirstInBucket = 0;
   statLastInBucket = 0;
   statMoveupHeap = 0;
   statMoveupSort = 0;
   statInsertHeap = 0;
   statInsertSort = 0;
   statStateChange = 0;
}

HOTQueue::HOTQueue(ulong minArcLen, ulong maxArcLen, ulong cLevels)
{
   ulong iBucket, logcLevelSize = 1, logSkip = 0;
   HOTLevel *pLevel;

   rgLevels = new HOTLevel[cLevels+1];    // one extra, maybe, for SUPER_TOP
   assert(rgLevels);
   topLevel = rgLevels + cLevels-1;
#ifndef NO_SUPER_TOP
   topLevel++;                     // this inludes an extra "super-top" level
#endif
   
       // we can have up to minarclen-1 different values at lowest level
   while ( (LOW_LEVEL_BUCKET_SIZE(minArcLen) >> logSkip) > 1 )
      logSkip++;                         // means 2^logSkip <= minArcLen

       // determine the number of buckets per level: it's about 
       // maxAL^(1/levels) / 2^rglogSkip[0].  *BUT* we insist it's a
       // power of two.  To avoid top-level nodes wrapping all the way
       // around, and for various theoretical reasons, we have a "super-
       // top" level with only two nodes, each of which subtends maxArcLen
       // distance values (or a power of 2 >= maxArcLen).
       // We store info in ulongs, so (# of buckets on level * skip) <= 32
   for ( logcLevelSize = 1;
	 (maxArcLen >> (logcLevelSize*cLevels + logSkip)) > 0;
	 logcLevelSize++ )
      if ( logcLevelSize*cLevels+1 + logSkip > sizeof(maxArcLen)*8 )
      {
	 fprintf(stderr, "FATAL ERROR: Maxarclen %ld is too long\n", maxArcLen);
	 exit(1);
      }
      // we expand buckets with more than minExpand items, and keep buckets
      // sorted until maxSort operations have been done
   minExpand = ((1<<logcLevelSize)/(logcLevelSize*cLevels))/4;
   maxSort = 5000;
   minHeapOrig = 2000;
   if ( minExpand < MIN_MIN_EXPAND )   // never expand nodes smaller than M_M_E
      minExpand = MIN_MIN_EXPAND;
   if ( minHeapOrig > minExpand )      // keep our constants in order
      minHeapOrig = minExpand;
   heapCurrent = new Heap(minExpand);
   assert(heapCurrent);

   for ( pLevel = rgLevels; pLevel <= topLevel; pLevel++ )
      pLevel->logcBuckets = logcLevelSize;
   topLevel->logcBuckets++;             // make sure we don't wrap into ourself
#ifndef NO_SUPER_TOP
   topLevel->logcBuckets = 1;           // override: "super-top" has 2 buckets
#endif
   rgLevels[0].logSkip = logSkip;       // the bootstrap value
   for ( pLevel = rgLevels; pLevel <= topLevel; pLevel++ )
   {
      pLevel->cBuckets = 1 << pLevel->logcBuckets;
      if ( pLevel != topLevel )        // note 0 has already been calculated
	 (pLevel+1)->logSkip = pLevel->logSkip + pLevel->logcBuckets;
      pLevel->distExtent = TWO_TO_X_LESS_1(pLevel->logSkip+pLevel->logcBuckets);
      pLevel->rgBin = new HOTBucket[pLevel->cBuckets];
      assert( pLevel->rgBin );
      for ( iBucket = 0; iBucket < pLevel->cBuckets; iBucket++ )
	 pLevel->rgBin[iBucket].pLevel = pLevel;   // store level info
   }

   Init();
}

HOTQueue::~HOTQueue()
{
   HOTLevel *pLevel;

   for ( pLevel = rgLevels; pLevel <= topLevel; pLevel++ )
      delete pLevel->rgBin;
   delete rgLevels;
   delete heapCurrent;
}

//------------------------------------------------------------
// HOTQueue::DistToBucket()
//      A node's distance (value) determines whether or not it
//      belongs in a given bucket.  Every node fits in some top
//      level bucket, but depending on what lower level bucket is
//      active, a given node may not fit in any bucket in a lower
//      level.  Given a node, we return the *lowest level bucket*
//      that the node belongs in, up to the level of bckCurrent.
//         It is possible, at the top level only, to wrap around,
//      This happens when node->distToBase > top-level extent.
//      We modify the distToBase in this case (which is why a pointer
//      to it is passed).  distToBase gives the distance to
//      the leftmost top-level bucket and uniquely determines the
//      bucket a node belongs in.  It allows us to stay a ulong.
//         Check that we're not at infinity before calling.
//         RETURNS: pointer to the bucket for a given value.
//------------------------------------------------------------

HOTBucket *HOTQueue::DistToBucket(ulong *pDistToBase)
{
   HOTLevel *pLevel;

      // A seeminly negative distance must be due to wrap-around
   if ( *pDistToBase < bckCurrent->pLevel->minDistToBase )
      return topLevel->rgBin + OFFSET(*pDistToBase, topLevel);

   for ( pLevel = bckCurrent->pLevel; pLevel <= topLevel; pLevel++ )
      if ( *pDistToBase <= pLevel->maxDistToBase )   // fits here
	 return pLevel->rgBin + OFFSET(*pDistToBase, pLevel);

      // For wrap-around, we only need to adjust DistToBase (take it mod)
   assert( *pDistToBase >= topLevel->distExtent+1 );   // not too small
   *pDistToBase -= topLevel->distExtent+1;             // get back in range
   assert( *pDistToBase <= topLevel->distExtent );     // not too big
   return topLevel->rgBin + OFFSET(*pDistToBase, topLevel);
}


//------------------------------------------------------------
// HOTQueue::HeapToExpand()
// HOTQueue::SortToExpand()
// HOTQueue::SortToHeap()
// HOTQueue::BckToHeap()
// HOTQueue::BckToSort()
//     Info can be stored in the current bucket in several ways.
//     These routines change from one way to the other.
//        NONE: nodes are in bckCurrent but shouldn't be.
//        SORT: nodes are stored between sortCurrent and 
//              sortCurrentLast.  Only NEXT() field is used.
//        HEAP: nodes are stored in heapCurrent.
//        EXPAND: nodes are stored in bckCurrent in an arbitrary
//                order, because RemoveMin will work by expanding
//                the current bucket.
//------------------------------------------------------------

void HOTQueue::HeapToExpand()
{
   Node *newNode;

   assert( stateCurrent == CURR_HEAP );
   assert( bckCurrent->pNode == NULL );
   while ( !heapCurrent->IsEmpty() )             // just move the nodes
   {
      newNode = heapCurrent->RemoveArbitrary();  // we don't care about order
      newNode->where = IN_BUCKETS;
      NEXT(newNode) = bckCurrent->pNode;         // don't need PREV to expand
      bckCurrent->pNode = newNode;
   }
   stateCurrent = CURR_BCK;
   STATE_CHANGE;
}

void HOTQueue::SortToExpand()
{
   assert( stateCurrent == CURR_SORT );
   assert( bckCurrent->pNode == NULL );
   bckCurrent->pNode = sortCurrent;  // there's nothing to do, really...
   sortCurrent = NULL;
   stateCurrent = CURR_BCK;          // just tell them not to bother sorting
   STATE_CHANGE;
}

void HOTQueue::SortToHeap()
{
   Node *nextNode;

   assert( stateCurrent == CURR_SORT );
   assert( heapCurrent->IsEmpty() );
   for ( ; sortCurrent != sortCurrentLast; sortCurrent = nextNode )
   {
      nextNode = NEXT(sortCurrent);
      heapCurrent->Insert(sortCurrent);   // note sort order is also heap order
   }
   heapCurrent->Insert(sortCurrent);      // insert last guy
   sortCurrent = NULL;
   stateCurrent = CURR_HEAP;
   STATE_CHANGE;
}

void HOTQueue::BckToHeap()
{
   Node *nextNode;

   assert( stateCurrent == CURR_BCK );
   assert( heapCurrent->IsEmpty() );
   for ( ; bckCurrent->pNode; bckCurrent->pNode = nextNode )
   {
      nextNode = NEXT(bckCurrent->pNode);
      heapCurrent->Insert(bckCurrent->pNode);
   }
   heapCurrent->Heapify();                // not already in heap order
   stateCurrent = CURR_HEAP;
   STATE_CHANGE;
}

void HOTQueue::BckToSort()
{
   Node *currNode = bckCurrent->pNode, *nextNode;

   assert( stateCurrent == CURR_BCK );
   assert( sortCurrent == NULL );
   stateCurrent = CURR_SORT;
   sortCurrent = sortCurrentLast = bckCurrent->pNode;
   bckCurrent->pNode = NULL;          // indicate a move to sortedness
   bckCurrent->cNodes = 1;            // InsertIntoCurrent() will change that
   for ( currNode = NEXT(sortCurrent); currNode; currNode = nextNode )
   {
      nextNode = NEXT(currNode);
      InsertIntoCurrent(currNode);
      assert( stateCurrent == CURR_SORT );    // did Insert() change it?
      bckCurrent->pLevel->cNodes--;   // incorrectly updated in Insert()
      assert( bckCurrent->pLevel->cNodes >= bckCurrent->cNodes );
   }
   STATE_CHANGE;
}


//------------------------------------------------------------
// HOTQueue::InsertIntoCurrent()
// HOTQueue::MoveUpInCurrent()
// HOTQueue::Insert()
// HOTQueue::MoveUp()
// HOTQueue::DeleteFromCurrent()
//     Insert() adds a node to a given bucket; MoveUp() deletes
//     the node from its current bucket and moves it to a new one
//     based on its new, updated-somehow distance.
//     Both have special semantics if bckNew is bckCurrent, but
//     normally we update level statistics.
//     All four functions return the node passed to them.
//        DeleteFromCurrent() deletes and RETURNS the smallest
//     node from bckCurrent, which mustn't be empty.
//------------------------------------------------------------

Node *HOTQueue::InsertIntoCurrent(Node *node)
{
   Node *nodePlace = sortCurrent;

   bckCurrent->cNodes++;
   BUCKET(node) = bckCurrent;           // in case we move it up some more
   node->where = IN_BUCKETS;            // this may change later
   bckCurrent->pLevel->cNodes++;

   if ( bckCurrent->cNodes == 1 )       // we were empty
   {
      assert( bckCurrent->pNode == NULL );
      sortCurrent = sortCurrentLast = node;
      NEXT(node) = PREV(node) = NULL;
      stateCurrent = CURR_SORT;         // easy to claim sortedness
      return node;
   }               

      // figure out if we need to change storage types
   if ( bckCurrent->cNodes >= minExpand && stateCurrent == CURR_HEAP )
      HeapToExpand();
   else if ( bckCurrent->cNodes >= minExpand && stateCurrent == CURR_SORT )
      SortToExpand();
   else if ( iSort > maxSort && stateCurrent == CURR_SORT )
      SortToHeap();

   switch ( stateCurrent )              // in sorted list, heap, or buckets?
   {
      case CURR_BCK:  NEXT(node) = bckCurrent->pNode;
                      bckCurrent->pNode = node;
                      break;
      case CURR_HEAP: heapCurrent->Insert(node);
                      heapCurrent->MoveUp(node);
                      INSERT_HEAP;                     // keep track of stats
                      break;
      case CURR_SORT: if (DIST_TO_BASE(node) >= DIST_TO_BASE(sortCurrentLast))
                      {                                // goes after end
                         NEXT(sortCurrentLast) = node;
			 PREV(node) = sortCurrentLast;
			 NEXT(node) = NULL;
			 sortCurrentLast = node;
                      }
                      else if (DIST_TO_BASE(node) <= DIST_TO_BASE(sortCurrent))
		      {                                // goes before beginning
			 NEXT(node) = sortCurrent;
			 PREV(sortCurrent) = node;
			 PREV(node) = NULL;
			 sortCurrent = node;
		      }
                      else
		      {                                // goes in middle
			 while (DIST_TO_BASE(node) > DIST_TO_BASE(NEXT(nodePlace)))
			 {
			    nodePlace = NEXT(nodePlace);
			    iSort++;                  // how much do we do?
			 }
			 NEXT(node) = NEXT(nodePlace);    // insert after...
			 NEXT(nodePlace) = node;          // nodePlace
			 PREV(node) = nodePlace;
			 PREV(NEXT(node)) = node;
		      }
                      INSERT_SORT;
                      break;
      default:        assert(0);
   }
   return node;
}  

Node *HOTQueue::MoveUpInCurrent(Node *node)
{
   Node *prevNew = sortCurrent;

   switch ( stateCurrent )              // in sorted list, heap, or buckets?
   {
      case CURR_BCK:  break;            // unordered to begin with?  no change
      case CURR_HEAP: heapCurrent->MoveUp(node);
                      MOVEUP_HEAP;
                      break;
      case CURR_SORT: if ( node == sortCurrent )  // node is already lowest
	                 break;                   // don't need to move up
                      iSort++;                    // for check against sortCurr
                      if ( DIST_TO_BASE(node) < DIST_TO_BASE(sortCurrent) )
		      {                           // we go first
			 if ( node == sortCurrentLast )    // first remove NODE
			    sortCurrentLast = PREV(node);
			 else
			    PREV(NEXT(node)) = PREV(node);
			 NEXT(PREV(node)) = NEXT(node);
			 NEXT(node) = sortCurrent;         // now insert NODE
			 PREV(node) = NULL;
			 PREV(sortCurrent) = node;
			 sortCurrent = node;
			 MOVEUP_SORT;
			 break;
		      }
                      while (DIST_TO_BASE(NEXT(prevNew)) < DIST_TO_BASE(node))
		      {
			 iSort++;                 // how much do we do?
			 prevNew = NEXT(prevNew);
		      }
                      if ( prevNew != PREV(node) )  // node moves
		      {
			 assert( NEXT(prevNew) != sortCurrentLast );
			 if ( node == sortCurrentLast )     // don't forget!
			    sortCurrentLast = PREV(node);
			 else
			    PREV(NEXT(node)) = PREV(node);  // next exists
			 NEXT(PREV(node)) = NEXT(node);     // remove node
			 NEXT(node) = NEXT(prevNew);        // insert node
			 PREV(node) = prevNew;
			 NEXT(prevNew) = node;              // look around node
			 PREV(NEXT(node)) = node;
		      }
                      MOVEUP_SORT;
                      break;
      default:        assert(0);
   }
   return node;                         // no stat changed: node in same bck
}

Node *HOTQueue::Insert(Node *node, HOTBucket *bckNew)
{
   if ( bckNew == bckCurrent )     // the current bucket is special
      return InsertIntoCurrent(node);

   BUCKET(node) = bckNew;          // where we live, now
   if ( bckNew->cNodes == 0 )      // we're the first one in the bucket
   {
      bckNew->pNode = node;
      NEXT(node) = PREV(node) = NULL;
      FIRST_IN_BUCKET;
   }
   else                            // insert us in cycle, before first
   {
      PREV(bckNew->pNode) = node;
      NEXT(node) = bckNew->pNode;
      PREV(node) = NULL;
      bckNew->pNode = node;
   }

   bckNew->cNodes++;
   if ( bckNew->pLevel == topLevel && bckNew < topLevel->pBucket )
   {                                 // wrap around
      cNodesAfterWrap++;             // we adjust a special variable for wrap
      if ( bckNew < minAfterWrap )
	 minAfterWrap = bckNew;
   }
   else
      bckNew->pLevel->cNodes++;      // another not at this level
   node->where = IN_BUCKETS;

   return node;
}

Node *HOTQueue::MoveUp(Node *node, HOTBucket *bckNew)
{
   HOTBucket *bckOld = BUCKET(node);

   if ( InBucket(node) && bckOld == bckCurrent || 
        heapCurrent->InHeap(node) )  // a kludge
      return MoveUpInCurrent(node);  // the current bucket is special

   if ( InBucket(node) )             // delete from its old location
   {
      if ( bckOld == bckNew )
	 return node;                // the "new" bucket ain't so new
      assert( bckOld->cNodes > 0 );
      if ( bckOld->cNodes == 1 )     // we're the last one in the bucket
      {
	 bckOld->pNode = NULL;       // this is not strictly necessary
	 LAST_IN_BUCKET;             // count a statistic
      }
      else                           // delete the node
      {
	 if ( PREV(node) )
	    NEXT(PREV(node)) = NEXT(node);
	 if ( NEXT(node) )
	    PREV(NEXT(node)) = PREV(node);
	 if ( bckOld->pNode == node )
	    bckOld->pNode = NEXT(node);
      }
      bckOld->cNodes--;
      if ( bckOld->pLevel == topLevel && bckOld < topLevel->pBucket )
	 cNodesAfterWrap--;          // we adjust a special variable for wrap
      else
	 bckOld->pLevel->cNodes--;
   }
   return Insert(node, bckNew);      // put the node in its new place
}

Node *HOTQueue::DeleteFromCurrent()
{
   Node *retval = NULL;

   assert( bckCurrent->cNodes > 0 );
   switch ( stateCurrent )              // in sorted list or heap?
   {
      case CURR_BCK:  retval = bckCurrent->pNode;    // any node is ok
                      bckCurrent->pNode = NEXT(retval);
                      break;
      case CURR_HEAP: assert( !heapCurrent->IsEmpty() );
                      retval = heapCurrent->RemoveMin();
                      break;
      case CURR_SORT: assert( sortCurrent != NULL );
                      retval = sortCurrent;
                      sortCurrent = NEXT(sortCurrent);
                      if ( sortCurrent != NULL )
			 PREV(sortCurrent) = NULL;
                      break;
      default:        assert(0);
   }
   retval->where = IN_NONE;
   bckCurrent->cNodes--;                // update important stats
   bckCurrent->pLevel->cNodes--;
   if ( bckCurrent->cNodes == 0 )       // make sure we're empty
   {
      sortCurrent = sortCurrentLast = NULL;
      bckCurrent->pNode = NULL;
   }
   return retval;
}

//------------------------------------------------------------
// HOTQueue::UpdateCurrentBucket()
// HOTQueue::SetupCurrentBucket()
// HOTQueue::RemoveMin()
//        UpdateCurrentBucket() finds the next bucket with at 
//     least one node in it and returns that bucket.  bckCurrent
//     should be empty.  MODIFIES bckCurrent to by the next
//     non-empty bucket, or NULL if all buckets are empty.
//        SetupCurrentBucket() deals with the current bucket
//     depending on how many elements it has, setting up a
//     sorted list, a heap, or expanding.  Expanding may
//     further MODIFY bckCurrent.  RETURNS the minimum
//     node, which it finds in the process of setting up.
//        RemoveMin() actually removes the smallest element
//     from bckCurrent, depending on the storage type.
//------------------------------------------------------------

void HOTQueue::UpdateCurrentBucket()
{
   HOTLevel *pLevel;

   stateCurrent = CURR_BCK;             // data just stored in a bucket
   iSort = 0;                           // no ops on our new bucket
     // we can pre-increment because the current bucket at each level must
     // be empty (because of previous expansion, or because it's bckCurrent)
   for ( pLevel = bckCurrent->pLevel; pLevel <= topLevel; pLevel++ )
      if ( pLevel->cNodes )                              // an item lives here
      {
	 while ( (++(pLevel->pBucket))->cNodes == 0 )
	    EMPTY_BUCKET;                                // update StatEmpty
	 assert(pLevel->pBucket - pLevel->rgBin < pLevel->cBuckets);
	 bckCurrent = pLevel->pBucket;                   // we're at this level
	 return;
      }

      // didn't find anything at any level.  Now we must wrap around
   topLevel->pBucket = minAfterWrap;
   topLevel->cNodes = cNodesAfterWrap;
   minAfterWrap = topLevel->rgBin + topLevel->cBuckets;  // reinit wraps
   cNodesAfterWrap = 0;
   baseToZero += topLevel->distExtent+1;       // the base increases as we wrap

      // familiar looking: but no preincrement because minWrap might be full
   if ( topLevel->cNodes )
   {
      while ( ((topLevel->pBucket)++)->cNodes == 0 )
	 EMPTY_BUCKET;                          // update StatEmpty
      (topLevel->pBucket)--;                    // undo last ++
      assert( topLevel->pBucket - topLevel->rgBin < topLevel->cBuckets );
      bckCurrent = topLevel->pBucket;
      return;
   }
   bckCurrent = NULL;                           // never found anything
   return;
}

Node *HOTQueue::SetupCurrentBucket()
{
   Node *pNode, *nextNode, *minNode, *swapNode;     // nodes in our bucket
   HOTBucket *bckNew = NULL;            // the bucket we insert pNode into
   HOTLevel *pLevel, *pLowLevel;
   
   if ( bckCurrent->pLevel == rgLevels )     // all distances are the same
   {
      stateCurrent = CURR_BCK;               // so just keep things here
      return DeleteFromCurrent();
   }
   if ( bckCurrent->cNodes >= minExpand )    // expand the current bucket
   {
      pLevel = bckCurrent->pLevel;
      pLowLevel = pLevel-1;

      pLowLevel->minDistToBase = pLevel->minDistToBase +  // SET UP LOWER LEVEL
	 ( (pLevel->pBucket - pLevel->rgBin) << pLevel->logSkip );
      pLowLevel->maxDistToBase =
	 pLowLevel->minDistToBase + pLowLevel->distExtent;
      pLowLevel->pBucket = pLowLevel->rgBin + pLowLevel->cBuckets;  // a max
      assert( pLowLevel->cNodes == 0);       // lower level should be empty
   
      pNode = bckCurrent->pNode;             // the bucket we're sorting
      minNode = pNode;                       // holds the return value
      nextNode = NEXT(pNode);
      while ( --bckCurrent->cNodes )         // go through all nodes but retval
      {
	 pNode = nextNode;
	 nextNode = NEXT(pNode);             // do this before we insert
            // if our node is smallest, expand the used-to-be smallest instead
	 if ( DIST_TO_BASE(pNode) < DIST_TO_BASE(minNode) )   // a min-ner node
	    swapNode = minNode, minNode = pNode, pNode = swapNode;
	 bckNew = pLowLevel->rgBin + OFFSET(DIST_TO_BASE(pNode), pLowLevel);
	 assert( bckNew >= pLowLevel->rgBin && 
 		 bckNew < pLowLevel->rgBin+pLowLevel->cBuckets );
	 Insert(pNode, bckNew);              // Insert increases cNodes
	 EXPANDED_NODE;                      // we've moved another node
      }
//      assert( nextNode == NULL );            // end of the list
      bckCurrent->pNode = NULL;              // indicate old stuff is empty
      bckCurrent = pLowLevel->pBucket =      // min bucket on lower level
	 pLowLevel->rgBin + OFFSET(DIST_TO_BASE(minNode), pLowLevel);
      stateCurrent = CURR_BCK;               // still just in buckets
      assert( pLevel->cNodes >= pLowLevel->cNodes+1 );  // how many we moved
      pLevel->cNodes -= pLowLevel->cNodes+1;            // these nodes moved
      EXPANDED_BUCKET;                       // we expanded another bucket
      minNode->where = IN_NONE;              // no longer in buckets
      return minNode;                        // only guy not expanded
   }
   else if ( bckCurrent->cNodes >= minHeapOrig )
      BckToHeap();
   else
      BckToSort();
   return DeleteFromCurrent();
}

//------------------------------------------------------------
// HOTQueue::RemoveMin()
//     If bckCurrent is empty, we update bckCurrent to the next
//     non-empty bucket.  Then we check if bckCurrent is set
//     up properly given the number of items in it (ie should
//     we go from a sorted list to a heap, or from a heap to
//     expanding?).  If so, we remove the minimum element; if
//     not we do some processing before removing the min.
//     RETURNS: the node we've removed, or NULL if there are no
//     more nodes.
//------------------------------------------------------------

Node *HOTQueue::RemoveMin()
{
   if ( bckCurrent->cNodes == 0 )
      UpdateCurrentBucket();        // no nodes anywhere
   if ( bckCurrent == NULL )        // no nodes *anywhere*
      return NULL;
   if ( stateCurrent == CURR_BCK )  // in raw list-of-nodes form
      return SetupCurrentBucket();  // expands, or puts in heap, or whatever
   else
      return DeleteFromCurrent();
}

//------------------------------------------------------------
// HOTQueue::PrintStats()
//     Prints stats specific to hotqueue implementation.
//     This includes all the bucket stats, and also some
//     hotqueue-specific stuff dealing with the active bucket:
//     how many times we change state (sort<->heap<->expand),
//     how many times we move up in the current bucket, both
//     in heap form and sort form.
//------------------------------------------------------------

void HOTQueue::PrintStats()
{
#ifdef ALLSTATS
   printf("c Empty Buckets: %12ld       \n", statEmpty);
   printf("c Expanded Nodes: %11ld       Expanded Buckets: %7ld\n",
	  statExpandedNodes, statExpandedBuckets);
   printf("c First In Bucket: %10ld       Last in bucket: %9ld\n",
	  statFirstInBucket, statLastInBucket);
   printf("c State Changes: %12ld       CurrBucket ops: %9ld\n", 
	  statStateChange, 
	  statMoveupHeap + statMoveupSort + statInsertHeap + statInsertSort);
   printf("c Heap MoveUps: %13ld       Sort MoveUps: %11ld\n",
	  statMoveupHeap, statMoveupSort);
   printf("c Heap Inserts: %13ld       Sort Inserts: %11ld\n",
	  statInsertHeap, statInsertSort);
#endif
}

void HOTQueue::dijkstra(Node *source, SP *sp)
{
   Node *currentNode, *newNode;   // newNode is beyond our current range
   Arc *arc, *lastArc;            // last arc of the current node

   Init();                        // reset indices
   DIST_TO_BASE(source) = (ulong)source->dist;  // no problem with wraparound
   Insert(source, DistToBucket(&(DIST_TO_BASE(source))));

   while ( (currentNode = RemoveMin()) )
   {
      sp->cScans++;
            // arcs are stored sequentially.  Our last arc is
            // one before the first arc of the next node.
      lastArc = (currentNode + 1)->first - 1;
      for ( arc = currentNode->first; arc <= lastArc; arc++ )
      {
	 newNode = arc->head;                      // where our arc ends up
	 if ( currentNode->dist + arc->len < newNode->dist )
	 {
	    newNode->dist = currentNode->dist + arc->len; // we\'re shorter
	    DIST_TO_BASE(newNode) = DIST_TO_BASE(currentNode) +
	       (ulong)arc->len;
	    newNode->parent = currentNode;               // update sp tree
	    newNode->current = arc;                      // arcs and nodes
	    MoveUp( newNode, DistToBucket(&(DIST_TO_BASE(newNode))) );
	    sp->cUpdates++;
	 }
      }
   }
}
