/* hotqueue.h  by craig silverstein.  Re-Started 5 april 1996.
 *     A new idea from Andrew Goldberg and others whose names I don't
 *     know.  We combine the best of buckets and heaps by using
 *     buckets when there are lots of nodes at a given level, and
 *     heaps when there are only a few.  Actually, instead of heaps we
 *     use a sorted list.  That means we have a bucket-like structure,
 *     except we do linear search instead of an expand-bucket if the
 *     high-level bucket has few items in it.
 */

#ifndef HOTQUEUE_H
#define HOTQUEUE_H

#include "nodearc.h"         // dfn of node and arc

#define CURR_NONE      0     // current bucket hasn't been dealt with yet
#define CURR_SORT      1     // current bucket is stored in a sorted list
#define CURR_HEAP      2     // current bucket is stored in a heap
#define CURR_BCK       3     // current bucket is just stored in a bucket

#ifndef ulong
typedef unsigned long ulong; // to get that extra bit
#endif

class SP;                // defined in sp.h
class Stack;             // defined in stack.h
class Heap;              // defined in heap.h
struct HOTLevel;         // defined below; defined here for Bucket

typedef struct HOTBucket {
   Node *pNode;          // pointer to the list of nodes at this level
   ulong cNodes;         // number of nodes in this bucket
   HOTLevel *pLevel;     // what level this bucket is on
} HOTBucket;

typedef struct HOTLevel {
   ulong cNodes;         // number of nodes currently stored at this level
   ulong cBuckets;       // number of buckets that exist on this level
   ulong logcBuckets;    // log (base 2) of the number of buckets
   ulong logSkip;        // how many buckets a bucket at this level subtends
   ulong distExtent;     // range of distance labels that will fit on our level
   ulong minDistToBase;  // distance from our level start to top level start
   ulong maxDistToBase;  // minDistToBase + distExtent
   HOTBucket *rgBin;     // points to the cBuckets buckets at our level
   HOTBucket *pBucket;   // the current bucket at level i
} HOTLevel;

class HOTQueue {
 private:
   HOTLevel *rgLevels;       // stats information about each level
   HOTLevel *topLevel;       // sentinel (last entry) for rgLevels
   double baseToZero;        // a measure of wraparound
   ulong cNodesAfterWrap;    // "Level" info for after we wrap around
   HOTBucket *minAfterWrap;
   ulong maxSort;            // we sort buckets until this many operations
   ulong iSort;              // number of operations we've seen so far
   ulong minHeapOrig;        // when first expanding, the sort/heap cutoff
   ulong minExpand;          // we heapify buckets until they're larger
   int stateCurrent;         // one of CURR_SORT, CURR_HEAP, CURR_BCK
   Node *sortCurrent;        // list of current-bucket contents in sorted order
   Node *sortCurrentLast;    // the last node in the sorted-order current bck
   Heap *heapCurrent;        // heap of current-bucket contents
   HOTBucket *bckCurrent;    // the currently active bucket
   long statEmpty;           // statistic: how many empty buckets we scanned
   long statExpandedNodes;   // statistic: how many nodes we expand out
   long statExpandedBuckets; // statistic: how many buckets we expand out of
   long statFirstInBucket;   // statistic: how many times we're node #1
   long statLastInBucket;    // statistic: how many times we're deleted last
   long statMoveupHeap;      // statistic: how many heap->Moveup() calls
   long statMoveupSort;      // statistic: how many moveups in sorted list
   long statInsertHeap;      // statistic: how many insertions into the heap
   long statInsertSort;      // statistic: how many insertions into sorted list
   long statStateChange;     // statistic: how many changes from sort/heap/exp

   HOTBucket *DistToBucket(ulong *pDistToBase);
   void HeapToExpand();
   void SortToExpand();
   void SortToHeap();
   void BckToHeap();
   void BckToSort();
   Node *MoveUpInCurrent(Node *node);
   Node *InsertIntoCurrent(Node *node);
   Node *DeleteFromCurrent();
   void UpdateCurrentBucket();
   Node *SetupCurrentBucket();
      
 public:
   HOTQueue(ulong minArcLen, ulong maxArcLen, ulong cLevels);
   ~HOTQueue();
   void Init();              // resets bucket indices.  Assumes buckets are OK
   void dijkstra(Node *source, SP *sp);            // run dijkstra's algorithm
   void PrintStats();

   Node *Insert(Node *node, HOTBucket *bckNew);
   Node *MoveUp(Node *node, HOTBucket *bckNew);
   Node *RemoveMin();
   int InBucket(Node *node)        { return node->where == IN_BUCKETS; }
};

#endif
