// sp.cc  by craig silverstein.  started 14 february 1995.
//     This is the sp type.  Only initialization code,
//     including a decision of what shortest path algorithm
//     to use, is included here.  Most of the code is in
//     dijkstra.cc, gor.cc, etc.  In all cases, SP computes
//     the shortest path in-place; the return value is only
//     important if the StopAt function is set, in which
//     case it\'s the vertex we stopped at.


#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include "sp.h"


double ArcLen(long cNodes, Node *nodes,
	      double *pMin /* = NULL */, double *pMax /* = NULL */)
// finds the max arc length and min arc length of a graph.
// Used to init for buckets.  pMax or pMin can be NULL.  For fun,
// and utility, returns MaxArcLen.
{
   Arc *lastArc, *arc;
   double maxLen = 0.0, minLen = VERY_FAR;

            // arcs are stored sequentially.  The last arc overall
            // is one before the first arc of the sentinel node.
   lastArc = (nodes+cNodes)->first - 1;
   for ( arc = nodes->first; arc <= lastArc; arc++ )
   {
      if ( arc->len > maxLen )
	 maxLen = arc->len;
      if ( arc->len < minLen )
         minLen = arc->len;
   }
   if ( pMin )   *pMin = minLen;
   if ( pMax )   *pMax = maxLen;
   return maxLen;
}


//-------------------------------------------------------------
// SP::SP()
// SP::~SP()
//     Each algorithm needs a different data structure.  We add
//     them all here and then send them in the sp() call.  This
//     way we need to allocate memory only once even though there
//     may be many sp calls.  Note that we expect the sp calls
//     all to be made on the same graph, though info about it
//     may change.
//-------------------------------------------------------------

SP::SP(long cNodesGiven, Node *nodesGiven, int spTypeGiven)
{
   cNodes = cNodesGiven;
   nodes = nodesGiven;
   spType = spTypeGiven;
   cCalls = cScans = cUpdates = 0;     // no stats yet
   heap = NULL;                        // for DIK_HEAP
   bucket = NULL;                      // for DIK_BUCKETS
   stkTopSort = stkSources = NULL;     // for GOR
   hotqueue = NULL;                    // for DIK_HOTQUEUE
   constbucket = NULL;                 // for DIK_CONSTBUCKETS
//** initialize data type for new sp algorithm here **//

   if ( spType & SP_DIK_BUCKETS )            // first because of its range
   {
      double minArcLen, maxArcLen;
      ArcLen(cNodes, nodes, &minArcLen, &maxArcLen);
      if ( spType == SP_DIK_BUCKETS )        // no # of buckets specified
	 spType = SP_DIK_BUCKETS + DIK_BUCKETS_DEFAULT;
      bucket = new BucketBin( (unsigned long)minArcLen,
			      (unsigned long)maxArcLen,
			      spType - SP_DIK_BUCKETS );
      spType = SP_DIK_BUCKETS;               // no need to store bucket #
   }

   else if ( spType & SP_DIK_HOTQUEUE )
   {
      double minArcLen, maxArcLen;
      ArcLen(cNodes, nodes, &minArcLen, &maxArcLen);
      if ( spType == SP_DIK_HOTQUEUE )       // no # of buckets specified
	 spType = SP_DIK_HOTQUEUE + DIK_BUCKETS_DEFAULT;
      hotqueue = new HOTQueue( (unsigned long)minArcLen,
			      (unsigned long)maxArcLen,
			      spType - SP_DIK_HOTQUEUE );
      spType = SP_DIK_HOTQUEUE;              // no need to store bucket #
   }

   else if ( spType & SP_DIK_HEAP )
      heap = new Heap(cNodes);

   else if ( spType & SP_GOR )
   {
      stkTopSort = new Stack(cNodes);
      stkSources = new Stack(cNodes);
   }

   else if ( spType & SP_DIK_CONSTBUCKETS )
   {
      double minArcLen, maxArcLen;
      ArcLen(cNodes, nodes, &minArcLen, &maxArcLen);
      constbucket = new ConstBucketBin( (unsigned long)minArcLen,
				        (unsigned long)maxArcLen );
   }

//** add initialization routine for new sp algorithm here **//

   else
      assert(0);             // crash on an illegal option
}

SP::~SP()
{
   if ( heap )        delete heap;
   if ( bucket )      delete bucket;
   if ( stkTopSort )  delete stkTopSort;
   if ( stkSources )  delete stkSources;
   if ( hotqueue )    delete hotqueue;
   if ( constbucket ) delete constbucket;
//** delete data structure for new sp algorithm here **//
}

//-------------------------------------------------------------
// SP::Init()
//     Initially, all lengths are set to VERY_FAR, presumably
//     longer than any real distance, and the parent field is
//     set to NULL to indicate the sp tree does not yet exist.
//     In addition, the temp field is set to an application-
//     specific (i.e. a parameter) NULL value.
//        The hard part is dealing with multiple sources.  We
//     can take either a single source or a stack of sources,
//     in either case we set their parent to themselves and
//     their distances to 0.
//-------------------------------------------------------------

void SP::Init(Node *source, Stack *stkSources, int spType)
{
   Node *currentNode;
   long iNode;             // to make our loop faster

   for ( currentNode=nodes, iNode=0; iNode <= cNodes; iNode++, currentNode++ )
   {
      currentNode->where = IN_NONE;   // nodes not in any data structure yet
      currentNode->dist = VERY_FAR;   // not yet a shortest path
      currentNode->parent = NULL;     // not yet in our sp tree
      switch ( spType ) {
	 case SP_DIK_HOTQUEUE:
	 case SP_DIK_HEAP:     break;   // don't need to do anything
	 case SP_DIK_BUCKETS:  currentNode->info.bckInfo.bucket = NULL; break;
	 case SP_DIK_CONSTBUCKETS: currentNode->info.constbckInfo.iLevel = -1;
                                   break;   // so it can't match in-bucket guy
	 case SP_GOR:          currentNode->info.status = 0;   break;
//** add initialization for new sp routine here **//
	 default: printf("shortest path type %d not supported\n", spType);
      }      
   }
   while ( source || (stkSources && (source=(Node *)stkSources->Iterate())) )
   {
      source->parent = source;
      source->dist = 0;               // all distances are to the source
      source = NULL;                  // so while can fail next time
   }
}

//-------------------------------------------------------------
// SP::sp()
//     Just calls the SP algorithm specified in spType, and returns
//     the overall stat that the SP algorithm returns.  Actual
//     code for sp algorithms can be found in their cc files:
//     dijkstra.cc, gor.cc, etc.
//        If OPEN is not NULL, we test all arcs with OPEN and
//     ignore those that aren\'t OPEN.  Otherwise all ARCS are open
//        If StopAt is not NULL, we do shortest paths only until
//     we reach a node for which StopAt returns 1 (it may check
//     for nodes with excess capacity, for instance).  Thus,
//     only part of the distance tree will be correct, but that
//     part will include a path from StopAt to a source.
//        RETURNS the node that StopAt stopped at, or NULL if
//     StopAt was NULL or no node passed StopAt\'s test.
//-------------------------------------------------------------

void SP::sp(Node *source)
{
   cCalls++;

   switch ( spType ) {
      case SP_DIK_HEAP: Init(source, stkSources, spType);
                        heap->dijkstra(source, this);
                        break;

      case SP_DIK_BUCKETS: Init(source, stkSources, spType);
   		           bucket->dijkstra(source, this);
                           break;

      case SP_GOR: Init(source, stkSources, spType);
                   gor(source);
                   break;

      case SP_DIK_HOTQUEUE: Init(source, stkSources, spType);
                            hotqueue->dijkstra(source, this);
                            break;

      case SP_DIK_CONSTBUCKETS: Init(source, stkSources, spType);
                                constbucket->dijkstra(source, this);
                                break;

//** add function call for new sp routine here **//
   }
}

//-------------------------------------------------------------
// SP::PrintStats()
//     Prints stats appropriate for an sp run.  First it prints
//     stats that hold for all sp algorithms: number of times
//     it's called, number of scans (nodes looked at), number
//     of updates (times a node's distance changes).  Then it
//     calls the appropriate data structure to print out
//     algorithm-specific stats.
//-------------------------------------------------------------

void SP::PrintStats()
{
   printf("c Scans: %20ld       Improvements: %11ld\n", cScans, cUpdates);

   switch ( spType ) {
      case SP_DIK_HEAP:         heap->PrintStats();         break;
      case SP_DIK_BUCKETS:      bucket->PrintStats();       break;
      case SP_GOR:                                          break;
      case SP_DIK_HOTQUEUE:     hotqueue->PrintStats();     break;
      case SP_DIK_CONSTBUCKETS: constbucket->PrintStats();  break;

//** add stats-printing function for new sp routine here **//

   }
}




