/* sp_run.cc  by craig silverstein.  started 23 february 1995.
 *     Driver for all the sp algorithms.  You call it like this:
 *         sp_run <sp-algorithm> < <datafile>
 *
 *     sp-algorithm can be one of SP_DIK_HEAP     1
 *                                SP_GOR          2
 *                                SP_DIK_BUCKETS  256 + #
 *     where # is the number of buckets you want.
 */

#include <stdlib.h>       // for atoi
#include <stdio.h>        // for printf
#include "sp.h"           // shortest-path class

extern int parse( long *n_ad, long *m_ad, Node **nodes_ad, Arc **arcs_ad, 
		  Node **source_ad, long *node_min_ad, char *problem_name );
extern double timer();            // in timer.cc: tells time use

#define ITERATIONS      1         // run the problem this many times
#define SZ_DIK_HEAP     "Dijkstra with heaps"
#define SZ_DIK_BUCKETS  "Dijkstra with buckets"
#define SZ_GOR          "GOR"
#define SZ_DIK_HOTQUEUE "Dijkstra with HOT Queues"
#define SZ_DIK_CONSTBUCKETS  "Dijkstra with constant-size buckets"

int main(int argc, char **argv)
{
   double tm;
   Arc *arcs;
   Node *nodes, *source, *node;
   long n, m, nmin, iIter;
   char name[21], *szAlgorithm;
   double dist = 0, maxArcLen, minArcLen;
   SP *sp;

   if ( argc < 2 )        // help
   {
      printf("Syntax: '%s <SP-algorithm>'.\n"
             "Choices are 1: Dijkstra heap algorithm\n"
             "            2: GOR algorithm\n"
             "            4: Dijkstra with constant-sized buckets\n"
             "            256+x: Dijkstra with x levels of buckets\n"
	     "            512+x: Dijkstra with x levels of buckets in HOT Queues\n",
	     argv[0]);
      return 1;     // 1: no argument
   }
   if ( atoi(argv[1]) == SP_DIK_HEAP )          szAlgorithm = SZ_DIK_HEAP; 
   else if ( atoi(argv[1]) == SP_GOR )          szAlgorithm = SZ_GOR;
   else if ( atoi(argv[1]) == SP_DIK_CONSTBUCKETS )
      szAlgorithm = SZ_DIK_CONSTBUCKETS;
   else if ( atoi(argv[1]) >= SP_DIK_BUCKETS &&
	     atoi(argv[1]) <  SP_DIK_HOTQUEUE ) szAlgorithm = SZ_DIK_BUCKETS;
   else if ( atoi(argv[1]) >= SP_DIK_HOTQUEUE ) szAlgorithm = SZ_DIK_HOTQUEUE;

   parse( &n, &m, &nodes, &arcs, &source, &nmin, name );   // in parse_sp.c
   sp = new SP(n, nodes, atoi(argv[1]));            

   tm = timer();                                           // in time.c
       // we only need to run sp once if we're just collecting stats
#ifndef ALLSTATS
   for ( iIter = 0; iIter < ITERATIONS; iIter++ )
#endif
      sp->sp(source);
   tm = (timer() - tm) / (double)ITERATIONS;               // give an avg time

   for ( node = nodes; node < nodes + n; node++ )          // calculate dists
      if ( node->parent )                                  // reachable
	 dist += node->dist;
   ArcLen(n, nodes, &minArcLen, &maxArcLen);      // other useful stats

   printf("c %s (%s) -> problem name: %s\n", szAlgorithm, argv[1], name);
   printf("c Nodes: %20ld       Arcs: %19ld\n",  n, m);
        /* *round* the time to the nearest .01 of a second */
   printf("c Time: %21.2f       MaxArcLen: %14.0f\n", tm + 0.005, maxArcLen);
   printf("c Cost: %21.0f       MinArcLen: %14.0f\n", dist, minArcLen);

      // now print the sp-specific stats
   sp->PrintStats();
   printf("\n");

   return 0;
}
